# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Bernoulli distribution."""

from typing import Any, Optional, Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Bernoulli(distribution.Distribution):
  """Bernoulli distribution.

  Bernoulli distribution with parameter `probs`, the probability of outcome `1`.
  """

  equiv_tfp_cls = tfd.Bernoulli

  def __init__(self,
               logits: Optional[Numeric] = None,
               probs: Optional[Numeric] = None,
               dtype: Union[jnp.dtype, type[Any]] = int):
    """Initializes a Bernoulli distribution.

    Args:
      logits: Logit transform of the probability of a `1` event (`0` otherwise),
        i.e. `probs = sigmoid(logits)`. Only one of `logits` or `probs` can be
        specified.
      probs: Probability of a `1` event (`0` otherwise). Only one of `logits` or
        `probs` can be specified.
      dtype: The type of event samples.
    """
    super().__init__()
    # Validate arguments.
    if (logits is None) == (probs is None):
      raise ValueError(
          f'One and exactly one of `logits` and `probs` should be `None`, '
          f'but `logits` is {logits} and `probs` is {probs}.')
    if not (jnp.issubdtype(dtype, bool) or
            jnp.issubdtype(dtype, jnp.integer) or
            jnp.issubdtype(dtype, jnp.floating)):
      raise ValueError(
          f'The dtype of `{self.name}` must be boolean, integer or '
          f'floating-point, instead got `{dtype}`.')
    # Parameters of the distribution.
    self._probs = None if probs is None else conversion.as_float_array(probs)
    self._logits = None if logits is None else conversion.as_float_array(logits)
    self._dtype = dtype

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """See `Distribution.event_shape`."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """See `Distribution.batch_shape`."""
    if self._logits is not None:
      return self._logits.shape
    return self._probs.shape

  @property
  def logits(self) -> Array:
    """The logits of a `1` event."""
    if self._logits is not None:
      return self._logits
    return jnp.log(self._probs) - jnp.log(1 - self._probs)

  @property
  def probs(self) -> Array:
    """The probabilities of a `1` event.."""
    if self._probs is not None:
      return self._probs
    return jax.nn.sigmoid(self._logits)

  def _log_probs_parameter(self) -> Tuple[Array, Array]:
    if self._logits is None:
      return (jnp.log1p(-1. * self._probs),
              jnp.log(self._probs))
    return (-jax.nn.softplus(self._logits),
            -jax.nn.softplus(-1. * self._logits))

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    probs = self.probs
    new_shape = (n,) + probs.shape
    uniform = jax.random.uniform(
        key=key, shape=new_shape, dtype=probs.dtype, minval=0., maxval=1.)
    return jnp.less(uniform, probs).astype(self._dtype)  # pylint: disable=not-callable

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    log_probs0, log_probs1 = self._log_probs_parameter()
    return (math.multiply_no_nan(log_probs0, 1 - value) +
            math.multiply_no_nan(log_probs1, value))

  def prob(self, value: EventT) -> Array:
    """See `Distribution.prob`."""
    probs1 = self.probs
    probs0 = 1 - probs1
    return (math.multiply_no_nan(probs0, 1 - value) +
            math.multiply_no_nan(probs1, value))

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    # For value < 0 the output should be zero because support = {0, 1}.
    return jnp.where(value < 0,
                     jnp.array(0., dtype=self.probs.dtype),
                     jnp.where(value >= 1,
                               jnp.array(1.0, dtype=self.probs.dtype),
                               1 - self.probs))

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def entropy(self) -> Array:
    """See `Distribution.entropy`."""
    (probs0, probs1,
     log_probs0, log_probs1) = _probs_and_log_probs(self)
    return -1. * (
        math.multiply_no_nan(log_probs0, probs0) +
        math.multiply_no_nan(log_probs1, probs1))

  def mean(self) -> Array:
    """See `Distribution.mean`."""
    return self.probs

  def variance(self) -> Array:
    """See `Distribution.variance`."""
    return (1 - self.probs) * self.probs

  def mode(self) -> Array:
    """See `Distribution.probs`."""
    return (self.probs > 0.5).astype(self._dtype)

  def __getitem__(self, index) -> 'Bernoulli':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    if self._logits is not None:
      return Bernoulli(logits=self.logits[index], dtype=self._dtype)
    return Bernoulli(probs=self.probs[index], dtype=self._dtype)


def _probs_and_log_probs(
    dist: Union[Bernoulli, tfd.Bernoulli]
    ) -> Tuple[Array, Array, Array, Array]:
  """Calculates both `probs` and `log_probs`."""
  # pylint: disable=protected-access
  if dist._logits is None:
    probs0 = 1. - dist._probs
    probs1 = 1. - probs0
    log_probs0 = jnp.log1p(-1. * dist._probs)
    log_probs1 = jnp.log(dist._probs)
  else:
    probs0 = jax.nn.sigmoid(-1. * dist._logits)
    probs1 = jax.nn.sigmoid(dist._logits)
    log_probs0 = -jax.nn.softplus(dist._logits)
    log_probs1 = -jax.nn.softplus(-1. * dist._logits)
  return probs0, probs1, log_probs0, log_probs1


def _kl_divergence_bernoulli_bernoulli(
    dist1: Union[Bernoulli, tfd.Bernoulli],
    dist2: Union[Bernoulli, tfd.Bernoulli],
    *unused_args, **unused_kwargs,
    ) -> Array:
  """KL divergence `KL(dist1 || dist2)` between two Bernoulli distributions.

  Args:
    dist1: instance of a Bernoulli distribution.
    dist2: instance of a Bernoulli distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  one_minus_p1, p1, log_one_minus_p1, log_p1 = _probs_and_log_probs(dist1)
  _, _, log_one_minus_p2, log_p2 = _probs_and_log_probs(dist2)
  # KL[a || b] = Pa * Log[Pa / Pb] + (1 - Pa) * Log[(1 - Pa) / (1 - Pb)]
  # Multiply each factor individually to avoid Inf - Inf
  return (
      math.multiply_no_nan(log_p1, p1) -
      math.multiply_no_nan(log_p2, p1) +
      math.multiply_no_nan(log_one_minus_p1, one_minus_p1) -
      math.multiply_no_nan(log_one_minus_p2, one_minus_p1)
  )


# Register the KL functions with TFP.
tfd.RegisterKL(Bernoulli, Bernoulli)(
    _kl_divergence_bernoulli_bernoulli)
tfd.RegisterKL(Bernoulli, Bernoulli.equiv_tfp_cls)(
    _kl_divergence_bernoulli_bernoulli)
tfd.RegisterKL(Bernoulli.equiv_tfp_cls, Bernoulli)(
    _kl_divergence_bernoulli_bernoulli)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""LU-decomposed affine bijector."""

from distrax._src.bijectors import bijector as base
from distrax._src.bijectors import block
from distrax._src.bijectors import chain
from distrax._src.bijectors import shift
from distrax._src.bijectors import triangular_linear
from distrax._src.bijectors import unconstrained_affine
import jax.numpy as jnp

Array = base.Array


class LowerUpperTriangularAffine(chain.Chain):
  """An affine bijector whose weight matrix is parameterized as A = LU.

  This bijector is defined as `f(x) = Ax + b` where:

  * A = LU is a DxD matrix.
  * L is a lower-triangular matrix with ones on the diagonal.
  * U is an upper-triangular matrix.

  The Jacobian determinant can be computed in O(D) as follows:

  log|det J(x)| = log|det A| = sum(log|diag(U)|)

  The inverse can be computed in O(D^2) by solving two triangular systems:

  * Lz = y - b
  * Ux = z

  The bijector is invertible if and only if all diagonal elements of U are
  non-zero. It is the responsibility of the user to make sure that this is the
  case; the class will make no attempt to verify that the bijector is
  invertible.

  L and U are parameterized using a square matrix M as follows:

  * The lower-triangular part of M (excluding the diagonal) becomes L.
  * The upper-triangular part of M (including the diagonal) becomes U.

  The parameterization is such that if M is the identity, LU is also the
  identity. Note however that M is not generally equal to LU.
  """

  def __init__(self, matrix: Array, bias: Array):
    """Initializes a `LowerUpperTriangularAffine` bijector.

    Args:
      matrix: a square matrix parameterizing `L` and `U` as described in the
        class docstring. Can also be a batch of matrices. If `matrix` is the
        identity, `LU` is also the identity. Note however that `matrix` is
        generally not equal to the product `LU`.
      bias: the vector `b` in `LUx + b`. Can also be a batch of vectors.
    """
    unconstrained_affine.check_affine_parameters(matrix, bias)
    self._upper = triangular_linear.TriangularLinear(matrix, is_lower=False)
    dim = matrix.shape[-1]
    lower = jnp.eye(dim) + jnp.tril(matrix, -1)  # Replace diagonal with ones.
    self._lower = triangular_linear.TriangularLinear(lower, is_lower=True)
    self._shift = block.Block(shift.Shift(bias), 1)
    self._bias = bias
    super().__init__([self._shift, self._lower, self._upper])

  @property
  def lower(self) -> Array:
    """The lower triangular matrix `L` with ones in the diagonal."""
    return self._lower.matrix

  @property
  def upper(self) -> Array:
    """The upper triangular matrix `U`."""
    return self._upper.matrix

  @property
  def matrix(self) -> Array:
    """The matrix `A = LU` of the transformation."""
    return self.lower @ self.upper

  @property
  def bias(self) -> Array:
    """The shift `b` of the transformation."""
    return self._bias

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is LowerUpperTriangularAffine:  # pylint: disable=unidiomatic-typecheck
      return all((
          self.lower is other.lower,
          self.upper is other.upper,
          self.bias is other.bias,
      ))
    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `diag_linear.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.bijectors.diag_linear import DiagLinear
from distrax._src.bijectors.tanh import Tanh
import haiku as hk
import jax
import jax.numpy as jnp
import numpy as np


class DiagLinearTest(parameterized.TestCase):

  def test_static_properties(self):
    bij = DiagLinear(diag=jnp.ones((4,)))
    self.assertTrue(bij.is_constant_jacobian)
    self.assertTrue(bij.is_constant_log_det)
    self.assertEqual(bij.event_ndims_in, 1)
    self.assertEqual(bij.event_ndims_out, 1)

  @parameterized.parameters(
      {'batch_shape': (), 'dtype': jnp.float16},
      {'batch_shape': (2, 3), 'dtype': jnp.float32},
  )
  def test_properties(self, batch_shape, dtype):
    bij = DiagLinear(diag=jnp.ones(batch_shape + (4,), dtype))
    self.assertEqual(bij.event_dims, 4)
    self.assertEqual(bij.batch_shape, batch_shape)
    self.assertEqual(bij.dtype, dtype)
    self.assertEqual(bij.diag.shape, batch_shape + (4,))
    self.assertEqual(bij.matrix.shape, batch_shape + (4, 4))
    self.assertEqual(bij.diag.dtype, dtype)
    self.assertEqual(bij.matrix.dtype, dtype)
    np.testing.assert_allclose(bij.diag, 1., atol=1e-6)
    np.testing.assert_allclose(
        bij.matrix, np.tile(np.eye(4), batch_shape + (1, 1)), atol=1e-6)

  def test_raises_with_invalid_parameters(self):
    with self.assertRaises(ValueError):
      DiagLinear(diag=np.ones(()))

  @chex.all_variants
  @parameterized.parameters(
      ((5,), (5,)),
      ((5,), ()),
      ((), (5,)),
  )
  def test_batched_parameters(self, diag_batch_shape, input_batch_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = jax.random.uniform(next(prng), diag_batch_shape + (4,)) + 0.5
    bij = DiagLinear(diag)

    x = jax.random.normal(next(prng), input_batch_shape + (4,))
    y, logdet_fwd = self.variant(bij.forward_and_log_det)(x)
    z, logdet_inv = self.variant(bij.inverse_and_log_det)(x)

    output_batch_shape = jnp.broadcast_shapes(
        diag_batch_shape, input_batch_shape)

    self.assertEqual(y.shape, output_batch_shape + (4,))
    self.assertEqual(z.shape, output_batch_shape + (4,))
    self.assertEqual(logdet_fwd.shape, output_batch_shape)
    self.assertEqual(logdet_inv.shape, output_batch_shape)

    diag = jnp.broadcast_to(diag, output_batch_shape + (4,)).reshape((-1, 4))
    x = jnp.broadcast_to(x, output_batch_shape + (4,)).reshape((-1, 4))
    y = y.reshape((-1, 4))
    z = z.reshape((-1, 4))
    logdet_fwd = logdet_fwd.flatten()
    logdet_inv = logdet_inv.flatten()

    for i in range(np.prod(output_batch_shape)):
      bij = DiagLinear(diag=diag[i])
      this_y, this_logdet_fwd = self.variant(bij.forward_and_log_det)(x[i])
      this_z, this_logdet_inv = self.variant(bij.inverse_and_log_det)(x[i])
      np.testing.assert_allclose(this_y, y[i], atol=1e-6)
      np.testing.assert_allclose(this_z, z[i], atol=1e-6)
      np.testing.assert_allclose(this_logdet_fwd, logdet_fwd[i], atol=1e-6)
      np.testing.assert_allclose(this_logdet_inv, logdet_inv[i], atol=1e-6)

  @chex.all_variants
  @parameterized.parameters(
      {'batch_shape': (), 'param_shape': ()},
      {'batch_shape': (2, 3), 'param_shape': (3,)},
  )
  def test_identity_initialization(self, batch_shape, param_shape):
    bij = DiagLinear(diag=jnp.ones(param_shape + (4,)))
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    x = jax.random.normal(next(prng), batch_shape + (4,))

    # Forward methods.
    y, logdet = self.variant(bij.forward_and_log_det)(x)
    np.testing.assert_array_equal(y, x)
    np.testing.assert_array_equal(logdet, jnp.zeros(batch_shape))

    # Inverse methods.
    x_rec, logdet = self.variant(bij.inverse_and_log_det)(y)
    np.testing.assert_array_equal(x_rec, y)
    np.testing.assert_array_equal(logdet, jnp.zeros(batch_shape))

  @chex.all_variants
  @parameterized.parameters(
      {'batch_shape': (), 'param_shape': ()},
      {'batch_shape': (2, 3), 'param_shape': (3,)}
  )
  def test_inverse_methods(self, batch_shape, param_shape):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = jax.random.uniform(next(prng), param_shape + (4,)) + 0.5
    bij = DiagLinear(diag)
    x = jax.random.normal(next(prng), batch_shape + (4,))
    y, logdet_fwd = self.variant(bij.forward_and_log_det)(x)
    x_rec, logdet_inv = self.variant(bij.inverse_and_log_det)(y)
    np.testing.assert_allclose(x_rec, x, atol=1e-6)
    np.testing.assert_allclose(logdet_fwd, -logdet_inv, atol=1e-6)

  @chex.all_variants
  def test_forward_jacobian_det(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = jax.random.uniform(next(prng), (4,)) + 0.5
    bij = DiagLinear(diag)

    batched_x = jax.random.normal(next(prng), (10, 4))
    single_x = jax.random.normal(next(prng), (4,))
    batched_logdet = self.variant(bij.forward_log_det_jacobian)(batched_x)

    jacobian_fn = jax.jacfwd(bij.forward)
    logdet_numerical = jnp.linalg.slogdet(jacobian_fn(single_x))[1]
    for logdet in batched_logdet:
      np.testing.assert_allclose(logdet, logdet_numerical, atol=5e-4)

  @chex.all_variants
  def test_inverse_jacobian_det(self):
    prng = hk.PRNGSequence(jax.random.PRNGKey(42))
    diag = jax.random.uniform(next(prng), (4,)) + 0.5
    bij = DiagLinear(diag)

    batched_y = jax.random.normal(next(prng), (10, 4))
    single_y = jax.random.normal(next(prng), (4,))
    batched_logdet = self.variant(bij.inverse_log_det_jacobian)(batched_y)

    jacobian_fn = jax.jacfwd(bij.inverse)
    logdet_numerical = jnp.linalg.slogdet(jacobian_fn(single_y))[1]
    for logdet in batched_logdet:
      np.testing.assert_allclose(logdet, logdet_numerical, atol=5e-4)

  def test_raises_on_invalid_input_shape(self):
    bij = DiagLinear(diag=jnp.ones((4,)))
    for fn_name, fn in [
        ('forward', bij.forward),
        ('inverse', bij.inverse),
        ('forward_log_det_jacobian', bij.forward_log_det_jacobian),
        ('inverse_log_det_jacobian', bij.inverse_log_det_jacobian),
        ('forward_and_log_det', bij.forward_and_log_det),
        ('inverse_and_log_det', bij.inverse_and_log_det)
    ]:
      with self.subTest(fn_name=fn_name):
        with self.assertRaises(ValueError):
          fn(jnp.array(0))

  def test_jittable(self):
    @jax.jit
    def f(x, b):
      return b.forward(x)

    bij = DiagLinear(diag=jnp.ones((4,)))
    x = np.zeros((4,))
    f(x, bij)

  def test_same_as_itself(self):
    bij = DiagLinear(diag=jnp.ones((4,)))
    self.assertTrue(bij.same_as(bij))

  def test_not_same_as_others(self):
    bij = DiagLinear(diag=jnp.ones((4,)))
    other = DiagLinear(diag=2. * jnp.ones((4,)))
    self.assertFalse(bij.same_as(other))
    self.assertFalse(bij.same_as(Tanh()))


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `math.py`."""

from absl.testing import absltest

from distrax._src.utils import math
import jax
import jax.numpy as jnp
import numpy as np
import scipy.special


class MathTest(absltest.TestCase):

  def test_multiply_no_nan(self):
    zero = jnp.zeros(())
    nan = zero / zero
    self.assertTrue(jnp.isnan(math.multiply_no_nan(zero, nan)))
    self.assertFalse(jnp.isnan(math.multiply_no_nan(nan, zero)))

  def test_multiply_no_nan_grads(self):
    x = -jnp.inf
    y = 0.
    self.assertEqual(math.multiply_no_nan(x, y), 0.)
    grad_fn = jax.grad(
        lambda inputs: math.multiply_no_nan(inputs[0], inputs[1]))
    np.testing.assert_allclose(grad_fn((x, y)), (y, x), rtol=1e-3)

  def test_power_no_nan(self):
    zero = jnp.zeros(())
    nan = zero / zero
    self.assertTrue(jnp.isnan(math.power_no_nan(zero, nan)))
    self.assertFalse(jnp.isnan(math.power_no_nan(nan, zero)))

  def test_power_no_nan_grads(self):
    x = np.exp(1.)
    y = 0.
    self.assertEqual(math.power_no_nan(x, y), 1.)
    grad_fn = jax.grad(
        lambda inputs: math.power_no_nan(inputs[0], inputs[1]))
    np.testing.assert_allclose(grad_fn((x, y)), (0., 1.), rtol=1e-3)

  def test_normalize_probs(self):
    pre_normalised_probs = jnp.array([0.4, 0.4, 0., 0.2])
    unnormalised_probs = jnp.array([4., 4., 0., 2.])
    expected_probs = jnp.array([0.4, 0.4, 0., 0.2])
    np.testing.assert_array_almost_equal(
        math.normalize(probs=pre_normalised_probs), expected_probs)
    np.testing.assert_array_almost_equal(
        math.normalize(probs=unnormalised_probs), expected_probs)

  def test_normalize_logits(self):
    unnormalised_logits = jnp.array([1., -1., 3.])
    expected_logits = jax.nn.log_softmax(unnormalised_logits, axis=-1)
    np.testing.assert_array_almost_equal(
        math.normalize(logits=unnormalised_logits), expected_logits)
    np.testing.assert_array_almost_equal(
        math.normalize(logits=expected_logits), expected_logits)

  def test_sum_last(self):
    x = jax.random.normal(jax.random.PRNGKey(42), (2, 3, 4))
    np.testing.assert_array_equal(math.sum_last(x, 0), x)
    np.testing.assert_array_equal(math.sum_last(x, 1), x.sum(-1))
    np.testing.assert_array_equal(math.sum_last(x, 2), x.sum((-2, -1)))
    np.testing.assert_array_equal(math.sum_last(x, 3), x.sum())

  def test_log_expbig_minus_expsmall(self):
    small = jax.random.normal(jax.random.PRNGKey(42), (2, 3, 4))
    big = small + jax.random.uniform(jax.random.PRNGKey(43), (2, 3, 4))
    expected_result = np.log(np.exp(big) - np.exp(small))
    np.testing.assert_allclose(
        math.log_expbig_minus_expsmall(big, small), expected_result, atol=1e-4)

  def test_log_beta(self):
    a = jnp.abs(jax.random.normal(jax.random.PRNGKey(42), (2, 3, 4)))
    b = jnp.abs(jax.random.normal(jax.random.PRNGKey(43), (3, 4)))
    expected_result = scipy.special.betaln(a, b)
    np.testing.assert_allclose(math.log_beta(a, b), expected_result, atol=2e-4)

  def test_log_beta_bivariate(self):
    a = jnp.abs(jax.random.normal(jax.random.PRNGKey(42), (4, 3, 2)))
    expected_result = scipy.special.betaln(a[..., 0], a[..., 1])
    np.testing.assert_allclose(
        math.log_beta_multivariate(a), expected_result, atol=2e-4)

  def test_log_beta_multivariate(self):
    a = jnp.abs(jax.random.normal(jax.random.PRNGKey(42), (2, 3, 4)))
    expected_result = (jnp.sum(scipy.special.gammaln(a), axis=-1)
                       - scipy.special.gammaln(jnp.sum(a, axis=-1)))
    np.testing.assert_allclose(
        math.log_beta_multivariate(a), expected_result, atol=1e-3)

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `distribution.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import distribution
import jax
import jax.numpy as jnp
import numpy as np


class DummyUnivariateDist(distribution.Distribution):
  """Dummy univariate distribution for testing."""

  def _sample_n(self, key, n):
    return jax.random.uniform(key, shape=(n,))

  def log_prob(self, value):
    """Log probability density/mass function."""

  @property
  def event_shape(self):
    """Shape of the events."""
    return np.shape([])


class DummyMultivariateDist(distribution.Distribution):
  """Dummy multivariate distribution for testing."""

  def __init__(self, dimension):
    super().__init__()
    self._dimension = dimension

  def _sample_n(self, key, n):
    return jax.random.uniform(key, shape=(n,) + self._dimension)

  def log_prob(self, value):
    """Log probability density/mass function."""

  @property
  def event_shape(self):
    """Shape of the events."""
    return (self._dimension,)


class DummyNestedDist(distribution.Distribution):
  """Dummy distribution with nested events for testing."""

  def __init__(self, batch_shape=()) -> None:
    self._batch_shape = batch_shape

  def _sample_n(self, key, n):
    return dict(
        foo=jax.random.uniform(key, shape=(n,) + self._batch_shape),
        bar=jax.random.uniform(key, shape=(n,) + self._batch_shape + (3,)))

  def log_prob(self, value):
    """Log probability density/mass function."""

  @property
  def event_shape(self):
    """Shape of the events."""
    return dict(foo=(), bar=(3,))


class DistributionTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.uni_dist = DummyUnivariateDist()

  @chex.all_variants(with_jit=False, with_device=False, with_pmap=False)
  @parameterized.named_parameters(
      ('0d input', 1, (1,)),
      ('0d np.int16 input', np.int16(2), (2,)),
      ('0d np.int32 input', np.int32(2), (2,)),
      ('0d np.int64 input', np.int64(2), (2,)),
      ('1d tuple input', (2,), (2,)),
      ('1d list input', [2], (2,)),
      ('1d tuple of np.int32 input', (np.int32(2),), (2,)),
      ('2d input', (2, 3), (2, 3)),
      ('3d input', (2, 3, 4), (2, 3, 4)))
  def test_sample_univariate_shape(self, shape, expected_shape):
    sample_fn = self.variant(
        lambda key: self.uni_dist.sample(seed=key, sample_shape=shape))
    samples = sample_fn(0)
    np.testing.assert_equal(samples.shape, expected_shape)

  @chex.all_variants(with_jit=False, with_device=False, with_pmap=False)
  @parameterized.named_parameters(
      ('0d input', (5,), 1, (1, 5)),
      ('0d np.int16 input', (5,), np.int16(1), (1, 5)),
      ('0d np.int32 input', (5,), np.int32(1), (1, 5)),
      ('0d np.int64 input', (5,), np.int64(1), (1, 5)),
      ('1d tuple input', (5,), (2,), (2, 5)),
      ('1d list input', (5,), [2], (2, 5)),
      ('1d tuple of np.int32 input', (5,), (np.int32(2),), (2, 5)),
      ('2d input', (4, 5), (2, 3), (2, 3, 4, 5)))
  def test_sample_multivariate_shape(self, var_dim, shape, expected_shape):
    mult_dist = DummyMultivariateDist(var_dim)
    sample_fn = self.variant(
        lambda key: mult_dist.sample(seed=key, sample_shape=shape))
    samples = sample_fn(0)
    np.testing.assert_equal(samples.shape, expected_shape)

  @chex.all_variants(with_jit=False, with_device=False, with_pmap=False)
  @parameterized.named_parameters(
      ('0d input', 1),
      ('0d np.int16 input', np.int16(1)),
      ('0d np.int32 input', np.int32(1)),
      ('0d np.int64 input', np.int64(1)),
      ('1d tuple input', (2,)),
      ('1d list input', [2]),
      ('2d input', (2, 3)),
  )
  def test_sample_nested_shape(self, shape):
    dist = DummyNestedDist()
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=shape))
    samples = sample_fn(0)
    # Ensure shape is a tuple.
    try:
      iter(shape)
    except TypeError:
      shape = (shape,)
    shape = tuple(shape)
    np.testing.assert_equal(samples['foo'].shape, shape)
    np.testing.assert_equal(samples['bar'].shape, shape + (3,))

  @parameterized.named_parameters(
      ('empty batch', ()),
      ('1d batch', (3,)),
      ('2d batch', (3, 4)),
  )
  def test_nested_batch_shape(self, batch_shape):
    dist = DummyNestedDist(batch_shape=batch_shape)
    np.testing.assert_equal(dist.batch_shape, batch_shape)

  @chex.all_variants(with_jit=False, with_device=False, with_pmap=False)
  def test_sample_keys(self):
    shape = 5
    key = 0
    sample_fn = self.variant(
        lambda key: self.uni_dist.sample(seed=key, sample_shape=shape))
    samples_from_int = sample_fn(key)
    rng = jax.random.PRNGKey(key)
    samples_from_prngkey = sample_fn(rng)
    np.testing.assert_array_equal(samples_from_int, samples_from_prngkey)

  def test_jittable(self):

    @jax.jit
    def sampler(dist, seed):
      return dist.sample(seed=seed)

    seed = jax.random.PRNGKey(0)
    dist = DummyMultivariateDist((5,))
    np.testing.assert_array_equal(
        sampler(dist, seed=seed), dist.sample(seed=seed))

  @parameterized.named_parameters(
      ('int', int),
      ('np.int16', np.int16),
      ('np.int32', np.int32),
      ('np.int64', np.int64),
      ('PRNGKey', jax.random.PRNGKey),
  )
  def test_convert_seed(self, dtype):
    rng, _ = distribution.convert_seed_and_sample_shape(dtype(0), 2)
    jax.random.split(rng)  # Should not raise an error.

  @parameterized.named_parameters(
      ('int', 2, (2,)),
      ('np.int16', np.int16(2), (2,)),
      ('np.int32', np.int32(2), (2,)),
      ('np.int64', np.int64(2), (2,)),
      ('int tuple', (2, 3), (2, 3)),
      ('np.int16 tuple', (np.int16(2), np.int16(3)), (2, 3)),
      ('np.int32 tuple', (np.int32(2), np.int32(3)), (2, 3)),
      ('np.int64 tuple', (np.int64(2), np.int64(3)), (2, 3)),
      ('int list', [2, 3], (2, 3)),
      ('np.int16 list', [np.int16(2), np.int16(3)], (2, 3)),
      ('np.int32 list', [np.int32(2), np.int32(3)], (2, 3)),
      ('np.int64 list', [np.int64(2), np.int64(3)], (2, 3)),
  )
  def test_convert_sample_shape(self, shape_in, shape_out):
    _, sample_shape = distribution.convert_seed_and_sample_shape(0, shape_in)
    assert sample_shape == shape_out

  @parameterized.named_parameters(
      ('single', 0, (np.array([[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]),
                     np.array([[0, 0, 0, 0], [1, 1, 1, 1], [2, 2, 2, 2]]),
                     np.array([[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]]))),
      ('range', slice(-1),
       (np.array([[[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]]
                 ]), np.array([[[0, 0, 0, 0], [1, 1, 1, 1], [2, 2, 2, 2]]]),
        np.array([[[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]]]))),
      ('two_axes', (slice(None), 1), (np.array(
          [[0, 0, 0, 0], [1, 1, 1, 1]]), np.array([[1, 1, 1, 1], [1, 1, 1, 1]]),
                                      np.array([[0, 1, 2, 3], [0, 1, 2, 3]]))),
      ('ellipsis', (Ellipsis, 2),
       (np.array([[0, 0, 0], [1, 1, 1]]), np.array(
           [[0, 1, 2], [0, 1, 2]]), np.array([[2, 2, 2], [2, 2, 2]]))),
      ('np_array', np.array([0, 1, -1]),
       ((np.array([[[0, 0, 0, 0], [0, 0, 0, 0], [0, 0, 0, 0]],
                   [[1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]],
                   [[1, 1, 1, 1], [1, 1, 1, 1], [1, 1, 1, 1]]],
                  dtype=np.int32),
         np.array([[[0, 0, 0, 0], [1, 1, 1, 1], [2, 2, 2, 2]],
                   [[0, 0, 0, 0], [1, 1, 1, 1], [2, 2, 2, 2]],
                   [[0, 0, 0, 0], [1, 1, 1, 1], [2, 2, 2, 2]]],
                  dtype=np.int32),
         np.array([[[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]],
                   [[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]],
                   [[0, 1, 2, 3], [0, 1, 2, 3], [0, 1, 2, 3]]],
                  dtype=np.int32)))),
  )
  def test_to_batch_shape_index(self, index, expected):
    np.testing.assert_allclose(
        distribution.to_batch_shape_index(batch_shape=(2, 3, 4), index=index),
        expected, 1e-3)

  def test_to_batch_shape_index_jnp_array(self):
    # This test needs to be a separate function since JAX doesn't allow creating
    # jnp.arrays in the top level of the program.
    # NOTE: Using jnp.arrays behaves differently compared to np.arrays as it
    # wraps instead of raising. Raising for same index is tested for np.arrays
    # below.
    index = (-1, 0)
    expected = (jnp.array([1, 1, 1, 1], dtype=jnp.int32),
                jnp.array([0, 0, 0, 0], dtype=jnp.int32),
                jnp.array([0, 1, 2, 3], dtype=jnp.int32))
    np.testing.assert_allclose(
        distribution.to_batch_shape_index(batch_shape=(2, 3, 4), index=index),
        expected, 1e-3)

  @parameterized.named_parameters(
      ('long_index', (1, 2, 3, 4)),
      ('np_array_out_of_bounds', np.array([-1, 2])),
  )
  def test_to_batch_shape_index_raises(self, index):
    with self.assertRaisesRegex(IndexError, 'not compatible with index'):
      distribution.to_batch_shape_index(
          batch_shape=(2, 3, 4), index=index)

  def test_multivariate_survival_function_raises(self):
    mult_dist = DummyMultivariateDist(42)
    with self.assertRaises(NotImplementedError):
      mult_dist.survival_function(jnp.zeros(42))
    with self.assertRaises(NotImplementedError):
      mult_dist.log_survival_function(jnp.zeros(42))

if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Distrax joint distribution over a tree of distributions."""

from typing import Any, Callable, Tuple, TypeVar

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


DistributionT = TypeVar(
    'DistributionT', bound=distribution.NestedT[conversion.DistributionLike]
)


def _is_leaf_distribution(x: Any) -> bool:
  return isinstance(x, conversion.DistributionLike)


def _map_up_to_distribution(
    fn: Callable[..., Any], distributions: Any, *xs: Any
) -> Any:
  """Maps `fn` over `x` up to `tree` to determine leaves."""
  return jax.tree.map(fn, distributions, *xs, is_leaf=_is_leaf_distribution)


def _leaves_up_to_distribution(distributions: Any) -> Any:
  """Flattens `distributions` up to `tree` to determine leaves."""
  return jax.tree.leaves(distributions, is_leaf=_is_leaf_distribution)


class Joint(distribution.Distribution):
  """Joint distribution over a tree of statistically independent distributions.

  Samples from the Joint distribution take the form of a tree structure that
  matches the structure of the underlying distributions. Log-probabilities
  are summed over the tree.

  All distributions in the tree must have the same `batch_shape` in order for
  log-probabilities to be computed correctly and for the `batch_shape` of the
  Joint distribution to be correct.
  """

  def __init__(self, distributions: DistributionT):
    """Initializes a Joint distribution over a tree of distributions.

    Args:
      distributions: Tree of distributions that must have the same batch shape.
    """
    super().__init__()
    self._distributions: DistributionT = _map_up_to_distribution(
        conversion.as_distribution, distributions
    )
    self._distributions_treedef = jax.tree.structure(
        self._distributions,
        is_leaf=_is_leaf_distribution,
    )
    self._distributions_leaves = _leaves_up_to_distribution(self._distributions)

    batch_shape = None
    for path, dist in jax.tree.leaves_with_path(
        self._distributions,
        is_leaf=_is_leaf_distribution,
        is_leaf_takes_path=False,
    ):
      batch_shape = batch_shape or dist.batch_shape
      first_path = '.'.join(map(str, path))
      if dist.batch_shape != batch_shape:
        path = '.'.join(map(str, path))
        raise ValueError(
            'Joint distributions must have the same batch shape, but '
            f'distribution "{dist.name}" at location {path} had batch shape '
            f'{dist.batch_shape} which is not equal to the batch shape '
            f'{batch_shape} of the distribution at location {first_path}.'
        )

  def _sample_n(self, key: chex.PRNGKey, n: int) -> distribution.EventT:
    keys = list(jax.random.split(key, len(self._distributions_leaves)))
    keys = self._distributions_treedef.unflatten(keys)
    return _map_up_to_distribution(
        lambda d, k: d.sample(seed=k, sample_shape=n), self._distributions, keys
    )

  def _sample_n_and_log_prob(
      self, key: chex.PRNGKey, n: int
  ) -> Tuple[distribution.EventT, chex.Array]:
    keys = list(jax.random.split(key, len(self._distributions_leaves)))
    keys = self._distributions_treedef.unflatten(keys)
    samples_and_log_probs = _map_up_to_distribution(
        lambda d, k: d.sample_and_log_prob(seed=k, sample_shape=n),
        self._distributions,
        keys,
    )
    samples = _map_up_to_distribution(
        lambda _, p: p[0], self._distributions, samples_and_log_probs
    )
    log_probs = _map_up_to_distribution(
        lambda _, p: p[1], self._distributions, samples_and_log_probs
    )
    log_probs = jnp.stack(jax.tree.leaves(log_probs))
    log_probs = jnp.sum(log_probs, axis=0)
    return samples, log_probs

  def log_prob(self, value: distribution.EventT) -> chex.Array:
    """Compute the total log probability of the distributions in the tree."""
    log_probs = _map_up_to_distribution(
        lambda dist, value: dist.log_prob(value), self._distributions, value
    )
    log_probs = jnp.stack(jax.tree.leaves(log_probs))
    return jnp.sum(log_probs, axis=0)

  @property
  def distributions(self) -> DistributionT:
    return self._distributions

  @property
  def event_shape(self) -> distribution.ShapeT:
    return _map_up_to_distribution(
        lambda dist: dist.event_shape, self._distributions
    )

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    return self._distributions_leaves[0].batch_shape

  @property
  def dtype(self) -> distribution.DTypeT:
    return _map_up_to_distribution(lambda dist: dist.dtype, self._distributions)

  def entropy(self) -> chex.Array:
    return sum(dist.entropy() for dist in self._distributions_leaves)

  def log_cdf(self, value: distribution.EventT) -> chex.Array:
    return sum(
        dist.log_cdf(v)
        for dist, v in zip(self._distributions_leaves, jax.tree.leaves(value))
    )

  def mean(self) -> distribution.EventT:
    """Calculates the mean."""
    return _map_up_to_distribution(
        lambda dist: dist.mean(), self._distributions
    )

  def median(self) -> distribution.EventT:
    """Calculates the median."""
    return _map_up_to_distribution(
        lambda dist: dist.median(), self._distributions
    )

  def mode(self) -> distribution.EventT:
    """Calculates the mode."""
    return _map_up_to_distribution(
        lambda dist: dist.mode(), self._distributions
    )

  def __getitem__(self, index) -> 'Joint':
    """See `Distribution.__getitem__`."""
    return Joint(
        _map_up_to_distribution(lambda dist: dist[index], self._distributions)
    )


def _kl_divergence_joint_joint(
    dist1: Joint, dist2: Joint, *unused_args, **unused_kwargs
) -> chex.Array:
  """Calculates the KL divergence between two Joint distributions."""
  treedef1 = jax.tree.structure(
      dist1.distributions, is_leaf=_is_leaf_distribution
  )
  treedef2 = jax.tree.structure(
      dist2.distributions, is_leaf=_is_leaf_distribution
  )
  if treedef1 != treedef2:
    raise ValueError(
        'Joint distributions must have the same tree structure, but\n'
        f'{treedef1=}\n{treedef2=}.'
    )
  return sum(
      inner1.kl_divergence(inner2)
      for inner1, inner2 in zip(
          _leaves_up_to_distribution(dist1.distributions),
          _leaves_up_to_distribution(dist2.distributions),
      )
  )


tfp.distributions.RegisterKL(Joint, Joint)(_kl_divergence_joint_joint)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Distrax adapter for Bijectors from TensorFlow Probability."""

from typing import Callable, Tuple

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp


tfb = tfp.bijectors

Array = base.Array


class BijectorFromTFP(base.Bijector):
  """Wrapper around a TFP bijector that turns it into a Distrax bijector.

  TFP bijectors and Distrax bijectors have similar but not identical semantics,
  which makes them not directly compatible. This wrapper guarantees that the
  wrapepd TFP bijector fully satisfies the semantics of Distrax, which enables
  any TFP bijector to be used by Distrax without modification.
  """

  def __init__(self, tfp_bijector: tfb.Bijector):
    """Initializes a BijectorFromTFP.

    Args:
      tfp_bijector: TFP bijector to convert to Distrax bijector.
    """
    self._tfp_bijector = tfp_bijector
    super().__init__(
        event_ndims_in=tfp_bijector.forward_min_event_ndims,
        event_ndims_out=tfp_bijector.inverse_min_event_ndims,
        is_constant_jacobian=tfp_bijector.is_constant_jacobian)

  def __getattr__(self, name: str):
    return getattr(self._tfp_bijector, name)

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    return self._tfp_bijector.forward(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    return self._tfp_bijector.inverse(y)

  def _ensure_batch_shape(self,
                          logdet: Array,
                          event_ndims_out: int,
                          forward_fn: Callable[[Array], Array],
                          x: Array) -> Array:
    """Broadcasts logdet to the batch shape as required."""
    if self._tfp_bijector.is_constant_jacobian:
      # If the Jacobian is constant, TFP may return a log det that doesn't have
      # full batch shape, but is broadcastable to it. Distrax assumes that the
      # log det is always batch-shaped, so we broadcast.
      y_shape = jax.eval_shape(forward_fn, x).shape
      if event_ndims_out == 0:
        batch_shape = y_shape
      else:
        batch_shape = y_shape[:-event_ndims_out]
      logdet = jnp.broadcast_to(logdet, batch_shape)
    return logdet

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    logdet = self._tfp_bijector.forward_log_det_jacobian(x, self.event_ndims_in)
    logdet = self._ensure_batch_shape(
        logdet, self.event_ndims_out, self._tfp_bijector.forward, x)
    return logdet

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    logdet = self._tfp_bijector.inverse_log_det_jacobian(
        y, self.event_ndims_out)
    logdet = self._ensure_batch_shape(
        logdet, self.event_ndims_in, self._tfp_bijector.inverse, y)
    return logdet

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    y = self._tfp_bijector.forward(x)
    logdet = self._tfp_bijector.forward_log_det_jacobian(x, self.event_ndims_in)
    logdet = self._ensure_batch_shape(
        logdet, self.event_ndims_out, self._tfp_bijector.forward, x)
    return y, logdet

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    x = self._tfp_bijector.inverse(y)
    logdet = self._tfp_bijector.inverse_log_det_jacobian(
        y, self.event_ndims_out)
    logdet = self._ensure_batch_shape(
        logdet, self.event_ndims_in, self._tfp_bijector.inverse, y)
    return x, logdet

  @property
  def name(self) -> str:
    """Name of the bijector."""
    return self._tfp_bijector.name



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Distrax: Probability distributions in JAX."""

# Bijectors.
from distrax._src.bijectors.bijector import Bijector
from distrax._src.bijectors.bijector import BijectorLike
from distrax._src.bijectors.block import Block
from distrax._src.bijectors.chain import Chain
from distrax._src.bijectors.diag_linear import DiagLinear
from distrax._src.bijectors.diag_plus_low_rank_linear import DiagPlusLowRankLinear
from distrax._src.bijectors.gumbel_cdf import GumbelCDF
from distrax._src.bijectors.inverse import Inverse
from distrax._src.bijectors.lambda_bijector import Lambda
from distrax._src.bijectors.linear import Linear
from distrax._src.bijectors.lower_upper_triangular_affine import LowerUpperTriangularAffine
from distrax._src.bijectors.masked_coupling import MaskedCoupling
from distrax._src.bijectors.rational_quadratic_spline import RationalQuadraticSpline
from distrax._src.bijectors.scalar_affine import ScalarAffine
from distrax._src.bijectors.shift import Shift
from distrax._src.bijectors.sigmoid import Sigmoid
from distrax._src.bijectors.split_coupling import SplitCoupling
from distrax._src.bijectors.tanh import Tanh
from distrax._src.bijectors.triangular_linear import TriangularLinear
from distrax._src.bijectors.unconstrained_affine import UnconstrainedAffine

# Distributions.
from distrax._src.distributions.bernoulli import Bernoulli
from distrax._src.distributions.beta import Beta
from distrax._src.distributions.categorical import Categorical
from distrax._src.distributions.categorical_uniform import CategoricalUniform
from distrax._src.distributions.clipped import Clipped
from distrax._src.distributions.clipped import ClippedLogistic
from distrax._src.distributions.clipped import ClippedNormal
from distrax._src.distributions.deterministic import Deterministic
from distrax._src.distributions.dirichlet import Dirichlet
from distrax._src.distributions.distribution import Distribution
from distrax._src.distributions.distribution import DistributionLike
from distrax._src.distributions.epsilon_greedy import EpsilonGreedy
from distrax._src.distributions.gamma import Gamma
from distrax._src.distributions.greedy import Greedy
from distrax._src.distributions.gumbel import Gumbel
from distrax._src.distributions.independent import Independent
from distrax._src.distributions.joint import Joint
from distrax._src.distributions.laplace import Laplace
from distrax._src.distributions.log_stddev_normal import LogStddevNormal
from distrax._src.distributions.logistic import Logistic
from distrax._src.distributions.mixture_of_two import MixtureOfTwo
from distrax._src.distributions.mixture_same_family import MixtureSameFamily
from distrax._src.distributions.multinomial import Multinomial
from distrax._src.distributions.mvn_diag import MultivariateNormalDiag
from distrax._src.distributions.mvn_diag_plus_low_rank import MultivariateNormalDiagPlusLowRank
from distrax._src.distributions.mvn_from_bijector import MultivariateNormalFromBijector
from distrax._src.distributions.mvn_full_covariance import MultivariateNormalFullCovariance
from distrax._src.distributions.mvn_tri import MultivariateNormalTri
from distrax._src.distributions.normal import Normal
from distrax._src.distributions.one_hot_categorical import OneHotCategorical
from distrax._src.distributions.quantized import Quantized
from distrax._src.distributions.softmax import Softmax
from distrax._src.distributions.straight_through import straight_through_wrapper
from distrax._src.distributions.transformed import Transformed
from distrax._src.distributions.uniform import Uniform
from distrax._src.distributions.von_mises import VonMises

# Utilities.
from distrax._src.utils.conversion import as_bijector
from distrax._src.utils.conversion import as_distribution
from distrax._src.utils.conversion import to_tfp
from distrax._src.utils.hmm import HMM
from distrax._src.utils.importance_sampling import importance_sampling_ratios
from distrax._src.utils.math import multiply_no_nan
from distrax._src.utils.monte_carlo import estimate_kl_best_effort
from distrax._src.utils.monte_carlo import mc_estimate_kl
from distrax._src.utils.monte_carlo import mc_estimate_kl_with_reparameterized
from distrax._src.utils.monte_carlo import mc_estimate_mode
from distrax._src.utils.transformations import register_inverse

__version__ = "0.1.7"

__all__ = (
    "as_bijector",
    "as_distribution",
    "Bernoulli",
    "Beta",
    "Bijector",
    "BijectorLike",
    "Block",
    "Categorical",
    "CategoricalUniform",
    "Chain",
    "Clipped",
    "ClippedLogistic",
    "ClippedNormal",
    "Deterministic",
    "DiagLinear",
    "DiagPlusLowRankLinear",
    "Dirichlet",
    "Distribution",
    "DistributionLike",
    "EpsilonGreedy",
    "estimate_kl_best_effort",
    "Gamma",
    "Greedy",
    "Gumbel",
    "GumbelCDF",
    "HMM",
    "importance_sampling_ratios",
    "Independent",
    "Inverse",
    "Joint",
    "Lambda",
    "Laplace",
    "Linear",
    "Logistic",
    "LogStddevNormal",
    "LowerUpperTriangularAffine",
    "MaskedCoupling",
    "mc_estimate_kl",
    "mc_estimate_kl_with_reparameterized",
    "mc_estimate_mode",
    "MixtureOfTwo",
    "MixtureSameFamily",
    "Multinomial",
    "multiply_no_nan",
    "MultivariateNormalDiag",
    "MultivariateNormalDiagPlusLowRank",
    "MultivariateNormalFromBijector",
    "MultivariateNormalFullCovariance",
    "MultivariateNormalTri",
    "Normal",
    "OneHotCategorical",
    "Quantized",
    "RationalQuadraticSpline",
    "register_inverse",
    "ScalarAffine",
    "Shift",
    "Sigmoid",
    "Softmax",
    "SplitCoupling",
    "straight_through_wrapper",
    "Tanh",
    "to_tfp",
    "Transformed",
    "TriangularLinear",
    "UnconstrainedAffine",
    "Uniform",
    "VonMises",
)


#  _________________________________________
# / Please don't use symbols in `_src` they \
# \ are not part of the Distrax public API. /
#  -----------------------------------------
#         \   ^__^
#          \  (oo)\_______
#             (__)\       )\/\
#                 ||----w |
#                 ||     ||
#



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Triangular linear bijector."""

import functools
from typing import Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.bijectors import linear
import jax
import jax.numpy as jnp

Array = base.Array


def _triangular_logdet(matrix: Array) -> Array:
  """Computes the log absolute determinant of a triangular matrix."""
  return jnp.sum(jnp.log(jnp.abs(jnp.diag(matrix))))


def _forward_unbatched(x: Array, matrix: Array) -> Array:
  return matrix @ x


def _inverse_unbatched(y: Array, matrix: Array, is_lower: bool) -> Array:
  return jax.scipy.linalg.solve_triangular(matrix, y, lower=is_lower)


class TriangularLinear(linear.Linear):
  """A linear bijector whose weight matrix is triangular.

  The bijector is defined as `f(x) = Ax` where `A` is a DxD triangular matrix.

  The Jacobian determinant can be computed in O(D) as follows:

  log|det J(x)| = log|det A| = sum(log|diag(A)|)

  The inverse is computed in O(D^2) by solving the triangular system `Ax = y`.

  The bijector is invertible if and only if all diagonal elements of `A` are
  non-zero. It is the responsibility of the user to make sure that this is the
  case; the class will make no attempt to verify that the bijector is
  invertible.
  """

  def __init__(self, matrix: Array, is_lower: bool = True):
    """Initializes a `TriangularLinear` bijector.

    Args:
      matrix: a square matrix whose triangular part defines `A`. Can also be a
        batch of matrices. Whether `A` is the lower or upper triangular part of
        `matrix` is determined by `is_lower`.
      is_lower: if True, `A` is set to the lower triangular part of `matrix`. If
        False, `A` is set to the upper triangular part of `matrix`.
    """
    if matrix.ndim < 2:
      raise ValueError(f"`matrix` must have at least 2 dimensions, got"
                       f" {matrix.ndim}.")
    if matrix.shape[-2] != matrix.shape[-1]:
      raise ValueError(f"`matrix` must be square; instead, it has shape"
                       f" {matrix.shape[-2:]}.")
    super().__init__(
        event_dims=matrix.shape[-1],
        batch_shape=matrix.shape[:-2],
        dtype=matrix.dtype)
    self._matrix = jnp.tril(matrix) if is_lower else jnp.triu(matrix)
    self._is_lower = is_lower
    triangular_logdet = jnp.vectorize(_triangular_logdet, signature="(m,m)->()")
    self._logdet = triangular_logdet(self._matrix)

  @property
  def matrix(self) -> Array:
    """The triangular matrix `A` of the transformation."""
    return self._matrix

  @property
  def is_lower(self) -> bool:
    """True if `A` is lower triangular, False if upper triangular."""
    return self._is_lower

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    self._check_forward_input_shape(x)
    batched = jnp.vectorize(_forward_unbatched, signature="(m),(m,m)->(m)")
    return batched(x, self._matrix)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    batch_shape = jax.lax.broadcast_shapes(self.batch_shape, x.shape[:-1])
    return jnp.broadcast_to(self._logdet, batch_shape)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self.forward(x), self.forward_log_det_jacobian(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    self._check_inverse_input_shape(y)
    batched = jnp.vectorize(
        functools.partial(_inverse_unbatched, is_lower=self._is_lower),
        signature="(m),(m,m)->(m)")
    return batched(y, self._matrix)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    return -self.forward_log_det_jacobian(y)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self.inverse(y), self.inverse_log_det_jacobian(y)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is TriangularLinear:  # pylint: disable=unidiomatic-typecheck
      return all((
          self.matrix is other.matrix,
          self.is_lower is other.is_lower,
      ))
    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `distribution_from_tfp.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions.categorical import Categorical
from distrax._src.distributions.distribution import Distribution
from distrax._src.distributions.distribution_from_tfp import distribution_from_tfp
from distrax._src.distributions.mvn_diag import MultivariateNormalDiag
from distrax._src.distributions.normal import Normal
from distrax._src.distributions.transformed import Transformed
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp

tfb = tfp.bijectors
tfd = tfp.distributions


class DistributionFromTfpNormal(parameterized.TestCase):
  """Tests for normal distribution."""

  def setUp(self):
    super().setUp()
    self._sample_shape = (np.int32(10),)
    self._seed = 42
    self._key = jax.random.PRNGKey(self._seed)
    self.base_dist = tfd.Normal(loc=0., scale=1.)
    self.values = jnp.array([1., -1.])
    self.distrax_second_dist = Normal(loc=-1., scale=0.8)
    self.tfp_second_dist = tfd.Normal(loc=-1., scale=0.8)

  def assertion_fn(self, rtol):
    return lambda x, y: np.testing.assert_allclose(x, y, rtol=rtol)

  @property
  def wrapped_dist(self):
    return distribution_from_tfp(self.base_dist)

  def test_event_shape(self):
    chex.assert_equal(self.wrapped_dist.event_shape, self.base_dist.event_shape)

  def test_batch_shape(self):
    chex.assert_equal(self.wrapped_dist.batch_shape, self.base_dist.batch_shape)

  @chex.all_variants
  def test_sample_dtype(self):
    samples = self.variant(self.wrapped_dist.sample)(seed=self._key)
    self.assertEqual(self.wrapped_dist.dtype, samples.dtype)
    self.assertEqual(self.wrapped_dist.dtype, self.base_dist.dtype)

  @chex.all_variants
  def test_sample(self):
    def sample_fn(key):
      return self.wrapped_dist.sample(sample_shape=self._sample_shape, seed=key)
    self.assertion_fn(rtol=2e-4)(
        self.variant(sample_fn)(self._key),
        self.base_dist.sample(sample_shape=self._sample_shape, seed=self._key))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('mean', 'mean'),
      ('mode', 'mode'),
      ('median', 'median'),
      ('stddev', 'stddev'),
      ('variance', 'variance'),
      ('entropy', 'entropy'),
  )
  def test_method(self, method):
    self.variant(lambda: None)  # To avoid variants usage error.
    try:
      expected_result = getattr(self.base_dist, method)()
    except NotImplementedError:
      return
    except AttributeError:
      return
    result = getattr(self.wrapped_dist, method)()
    self.assertion_fn(rtol=2e-4)(result, expected_result)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('log_prob', 'log_prob'),
      ('prob', 'prob'),
      ('log_cdf', 'log_cdf'),
      ('cdf', 'cdf'),
  )
  def test_method_with_value(self, method):
    self.variant(lambda: None)  # To avoid variants usage error.

    if (isinstance(self.base_dist, tfd.Categorical) and
        method in ('cdf', 'log_cdf')):
      # TODO(budden): make .cdf() and .log_cdf() from tfp.Categorical jittable.
      return

    try:
      expected_result = getattr(self.base_dist, method)(self.values)
    except NotImplementedError:
      return
    except AttributeError:
      return
    result = self.variant(getattr(self.wrapped_dist, method))(self.values)
    self.assertion_fn(rtol=2e-4)(result, expected_result)

  @chex.all_variants
  def test_sample_and_log_prob(self):
    base_samples = self.base_dist.sample(
        sample_shape=self._sample_shape, seed=self._key)
    base_logprob = self.base_dist.log_prob(base_samples)

    def sample_fn(key):
      return self.wrapped_dist.sample_and_log_prob(
          sample_shape=self._sample_shape, seed=key)
    samples, log_prob = self.variant(sample_fn)(self._key)
    self.assertion_fn(rtol=2e-4)(samples, base_samples)
    self.assertion_fn(rtol=2e-4)(log_prob, base_logprob)

  @chex.all_variants
  @parameterized.named_parameters(
      ('kl_divergence', 'kl_divergence'),
      ('cross_entropy', 'cross_entropy'),
  )
  def test_with_two_distributions(self, method):
    """Test methods of the forms listed below.

      D(distrax_distrib || wrapped_distrib),
      D(wrapped_distrib || distrax_distrib),
      D(tfp_distrib || wrapped_distrib),
      D(wrapped_distrib || tfp_distrib).

    Args:
      method: the method name to be tested
    """
    try:
      expected_result1 = self.variant(
          getattr(self.tfp_second_dist, method))(self.base_distribution)
      expected_result2 = self.variant(
          getattr(self.base_distribution, method))(self.tfp_second_dist)
    except NotImplementedError:
      return
    except AttributeError:
      return
    distrax_result1 = self.variant(getattr(self.distrax_second_dist, method))(
        self.wrapped_dist)
    distrax_result2 = self.variant(getattr(self.wrapped_dist, method))(
        self.distrax_second_dist)
    tfp_result1 = self.variant(getattr(self.tfp_second_dist, method))(
        self.wrapped_dist)
    tfp_result2 = self.variant(getattr(self.wrapped_dist, method))(
        self.tfp_second_dist)
    self.assertion_fn(rtol=2e-4)(distrax_result1, expected_result1)
    self.assertion_fn(rtol=2e-4)(distrax_result2, expected_result2)
    self.assertion_fn(rtol=2e-4)(tfp_result1, expected_result1)
    self.assertion_fn(rtol=2e-4)(tfp_result2, expected_result2)


class DistributionFromTfpMvnNormal(DistributionFromTfpNormal):
  """Tests for multivariate normal distribution."""

  def setUp(self):
    super().setUp()
    self.base_dist = tfd.MultivariateNormalDiag(loc=[0., 1.])
    self.values = jnp.array([1., -1.])
    self.distrax_second_dist = MultivariateNormalDiag(
        loc=jnp.array([-1., 0.]), scale_diag=jnp.array([0.8, 1.2]))
    self.tfp_second_dist = tfd.MultivariateNormalDiag(
        loc=[-1., 0.], scale_diag=[0.8, 1.2])

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))))
  def test_slice(self, slice_):
    loc = np.random.randn(3, 4, 5)
    scale_diag = np.random.randn(3, 4, 5)
    dist = distribution_from_tfp(
        tfd.MultivariateNormalDiag(loc=loc, scale_diag=scale_diag))
    sliced_dist = dist[slice_]
    self.assertIsInstance(sliced_dist, Distribution)
    self.assertion_fn(rtol=2e-4)(sliced_dist.mean(), loc[slice_])


class DistributionFromTfpCategorical(DistributionFromTfpNormal):
  """Tests for categorical distribution."""

  def setUp(self):
    super().setUp()
    self.base_dist = tfd.Categorical(logits=[0., -1., 1.])
    self.values = jnp.array([0, 1, 2])
    self.distrax_second_dist = Categorical(probs=jnp.array([0.2, 0.2, 0.6]))
    self.tfp_second_dist = tfd.Categorical(probs=[0.2, 0.2, 0.6])

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    logits = np.random.randn(3, 4, 5)
    probs = jax.nn.softmax(np.random.randn(3, 4, 5), axis=-1)
    dist1 = distribution_from_tfp(tfd.Categorical(logits=logits))
    dist2 = distribution_from_tfp(tfd.Categorical(probs=probs))
    sliced_dist1 = dist1[slice_]
    sliced_dist2 = dist2[slice_]
    self.assertIsInstance(sliced_dist1, Distribution)
    self.assertIsInstance(sliced_dist2, Distribution)
    self.assertIsInstance(sliced_dist1, tfd.Categorical)
    self.assertIsInstance(sliced_dist2, tfd.Categorical)
    self.assertion_fn(rtol=2e-4)(
        jax.nn.softmax(sliced_dist1.logits, axis=-1),
        jax.nn.softmax(logits[slice_], axis=-1))
    self.assertion_fn(rtol=2e-4)(sliced_dist2.probs, probs[slice_])

  def test_slice_ellipsis(self):
    logits = np.random.randn(3, 4, 5)
    probs = jax.nn.softmax(np.random.randn(3, 4, 5), axis=-1)
    dist1 = distribution_from_tfp(tfd.Categorical(logits=logits))
    dist2 = distribution_from_tfp(tfd.Categorical(probs=probs))
    sliced_dist1 = dist1[..., -1]
    sliced_dist2 = dist2[..., -1]
    self.assertIsInstance(sliced_dist1, Distribution)
    self.assertIsInstance(sliced_dist2, Distribution)
    self.assertIsInstance(sliced_dist1, tfd.Categorical)
    self.assertIsInstance(sliced_dist2, tfd.Categorical)
    self.assertion_fn(rtol=2e-4)(
        jax.nn.softmax(sliced_dist1.logits, axis=-1),
        jax.nn.softmax(logits[:, -1], axis=-1))
    self.assertion_fn(rtol=2e-4)(sliced_dist2.probs, probs[:, -1])


class DistributionFromTfpTransformed(DistributionFromTfpNormal):
  """Tests for transformed distributions."""

  def setUp(self):
    super().setUp()
    self.base_dist = tfd.TransformedDistribution(
        distribution=tfd.Normal(loc=0., scale=1.),
        bijector=tfb.Exp())
    self.values = jnp.array([0., 1., 2.])
    self.distrax_second_dist = Transformed(
        distribution=Normal(loc=0.5, scale=0.8),
        bijector=tfb.Exp())
    self.tfp_second_dist = tfd.TransformedDistribution(
        distribution=tfd.Normal(loc=0.5, scale=0.8),
        bijector=tfb.Exp())


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper for inverting a Distrax Bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.utils import conversion

Array = base.Array
BijectorLike = base.BijectorLike
BijectorT = base.BijectorT


class Inverse(base.Bijector):
  """A bijector that inverts a given bijector.

  That is, if `bijector` implements the transformation `f`, `Inverse(bijector)`
  implements the inverse transformation `f^{-1}`.

  The inversion is performed by swapping the forward with the corresponding
  inverse methods of the given bijector.
  """

  def __init__(self, bijector: BijectorLike):
    """Initializes an Inverse bijector.

    Args:
      bijector: the bijector to be inverted. It can be a distrax bijector, a TFP
        bijector, or a callable to be wrapped by `Lambda`.
    """
    self._bijector = conversion.as_bijector(bijector)
    super().__init__(
        event_ndims_in=self._bijector.event_ndims_out,
        event_ndims_out=self._bijector.event_ndims_in,
        is_constant_jacobian=self._bijector.is_constant_jacobian,
        is_constant_log_det=self._bijector.is_constant_log_det)

  @property
  def bijector(self) -> BijectorT:
    """The base bijector that was the input to `Inverse`."""
    return self._bijector

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    return self._bijector.inverse(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    return self._bijector.forward(y)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    return self._bijector.inverse_log_det_jacobian(x)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    return self._bijector.forward_log_det_jacobian(y)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self._bijector.inverse_and_log_det(x)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self._bijector.forward_and_log_det(y)

  @property
  def name(self) -> str:
    """Name of the bijector."""
    return self.__class__.__name__ + self._bijector.name

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is Inverse:  # pylint: disable=unidiomatic-typecheck
      return self.bijector.same_as(other.bijector)
    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Greedy distributions with respect to a set of preferences."""

from typing import Any, Union

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import distribution
import jax.numpy as jnp


Array = chex.Array


def _argmax_with_random_tie_breaking(preferences: Array) -> Array:
  """Compute probabilities greedily with respect to a set of preferences."""
  optimal_actions = preferences == preferences.max(axis=-1, keepdims=True)
  return optimal_actions / optimal_actions.sum(axis=-1, keepdims=True)


class Greedy(categorical.Categorical):
  """A Categorical distribution that is greedy with respect to some preferences.

  Given a set of unnormalized preferences, the probability mass is distributed
  equally among all indices `i` such that `preferences[i] = max(preferences)`,
  all other indices will be assigned a probability of zero.
  """

  def __init__(
      self, preferences: Array, dtype: Union[jnp.dtype, type[Any]] = int
  ):
    """Initializes a Greedy distribution.

    Args:
      preferences: Unnormalized preferences.
      dtype: The type of event samples.
    """
    self._preferences = jnp.asarray(preferences)
    probs = _argmax_with_random_tie_breaking(self._preferences)
    super().__init__(probs=probs, dtype=dtype)

  @property
  def preferences(self) -> Array:
    """Unnormalized preferences."""
    return self._preferences

  def __getitem__(self, index) -> 'Greedy':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Greedy(
        preferences=self.preferences[index], dtype=self.dtype)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `one_hot_categorical.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.distributions import categorical
from distrax._src.distributions import one_hot_categorical
from distrax._src.utils import equivalence
from distrax._src.utils import math
import jax
import jax.experimental
import jax.numpy as jnp
import numpy as np
import scipy
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions


class OneHotCategoricalTest(equivalence.EquivalenceTest):

  def setUp(self):
    super().setUp()
    self._init_distr_cls(one_hot_categorical.OneHotCategorical)
    self.p = np.asarray([0.1, 0.4, 0.2, 0.3])
    self.logits = np.log(self.p) - 1.0  # intended unnormalization

  def test_parameters_from_probs(self):
    dist = self.distrax_cls(probs=self.p)
    self.assertion_fn(rtol=2e-3)(
        dist.logits, math.normalize(logits=np.log(self.p)))
    self.assertion_fn(rtol=2e-3)(dist.probs, math.normalize(probs=self.p))

  def test_parameters_from_logits(self):
    dist = self.distrax_cls(logits=self.logits)
    self.assertion_fn(rtol=2e-3)(
        dist.logits, math.normalize(logits=self.logits))
    self.assertion_fn(rtol=2e-3)(dist.probs, math.normalize(probs=self.p))

  @parameterized.named_parameters(
      ('probs and logits', {'logits': [0.1, -0.2], 'probs': [0.6, 0.4]}),
      ('both probs and logits are None', {'logits': None, 'probs': None}),
      ('bool dtype', {'logits': [0.1, -0.2], 'dtype': jnp.bool_}),
      ('complex64 dtype', {'logits': [0.1, -0.2], 'dtype': jnp.complex64}),
      ('complex128 dtype', {'logits': [0.1, -0.2], 'dtype': jnp.complex128}),
  )
  def test_raises_on_invalid_inputs(self, dist_params):
    with self.assertRaises(ValueError):
      self.distrax_cls(**dist_params)

  @chex.all_variants
  def test_negative_probs(self):
    """Check sample returns -1 if probs are negative after normalization."""
    dist = self.distrax_cls(probs=np.asarray([[0.1, -0.4, 0.2, 0.3],
                                              [0.1, 0.1, 0.6, 0.2]]))
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=100))
    samples = sample_fn(self.key)
    self.assertion_fn(rtol=2e-3)(samples[:, 0, :], -1)
    np.testing.assert_array_compare(lambda x, y: x >= y, samples[:, 1, :], 0)

  @chex.all_variants
  def test_nan_probs(self):
    """Checks sample returns -1 if probs are nan after normalization."""
    dist = self.distrax_cls(
        probs=np.asarray([[-0.1, 0.1, 0.0, 0.0], [0.1, 0.1, 0.6, 0.2]]))
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=100))
    samples = sample_fn(self.key)
    self.assertion_fn(rtol=2e-3)(samples[:, 0, :], -1)
    np.testing.assert_array_compare(lambda x, y: x >= y, samples[:, 1, :], 0)

  @parameterized.named_parameters(
      ('from probs', False),
      ('from logits', True))
  def test_num_categories(self, from_logits):
    dist_params = {'logits': self.logits} if from_logits else {'probs': self.p}
    dist = self.distrax_cls(**dist_params)
    np.testing.assert_equal(dist.num_categories, len(self.p))

  @parameterized.named_parameters(
      ('1d logits', {'logits': [0.0, 1.0, -0.5]}),
      ('1d probs', {'probs': [0.2, 0.5, 0.3]}),
      ('2d logits', {'logits': [[0.0, 1.0, -0.5], [-0.1, 0.3, 0.0]]}),
      ('2d probs', {'probs': [[0.1, 0.4, 0.5], [0.5, 0.25, 0.25]]}),
  )
  def test_event_shape(self, distr_params):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_event_shape((), distr_params)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.2, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.2, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5],
                                          [-0.1, 0.3, 0.0]]}, ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5],
                                        [0.5, 0.25, 0.25]]}, ()),
      ('2d logits, int shape', {'logits': [[0.0, 1.0, -0.5],
                                           [-0.1, 0.3, 0.0]]}, 4),
      ('2d probs, int shape', {'probs': [[0.1, 0.4, 0.5],
                                         [0.5, 0.25, 0.25]]}, 4),
      ('2d logits, 1-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5,)),
      ('2d probs, 1-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5,)),
      ('2d logits, 2-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5, 4)),
  )
  def test_sample_shape(self, distr_params, sample_shape):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_sample_shape(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape)

  @chex.all_variants
  @parameterized.named_parameters(
      ('1d logits, no shape', {'logits': [0.0, 1.0, -0.5]}, ()),
      ('1d probs, no shape', {'probs': [0.2, 0.5, 0.3]}, ()),
      ('1d logits, int shape', {'logits': [0.0, 1.0, -0.5]}, 1),
      ('1d probs, int shape', {'probs': [0.2, 0.5, 0.3]}, 1),
      ('1d logits, 1-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (1,)),
      ('1d probs, 1-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (1,)),
      ('1d logits, 2-tuple shape', {'logits': [0.0, 1.0, -0.5]}, (5, 4)),
      ('1d probs, 2-tuple shape', {'probs': [0.2, 0.5, 0.3]}, (5, 4)),
      ('2d logits, no shape', {'logits': [[0.0, 1.0, -0.5],
                                          [-0.1, 0.3, 0.0]]}, ()),
      ('2d probs, no shape', {'probs': [[0.1, 0.4, 0.5],
                                        [0.5, 0.25, 0.25]]}, ()),
      ('2d logits, int shape', {'logits': [[0.0, 1.0, -0.5],
                                           [-0.1, 0.3, 0.0]]}, 4),
      ('2d probs, int shape', {'probs': [[0.1, 0.4, 0.5],
                                         [0.5, 0.25, 0.25]]}, 4),
      ('2d logits, 1-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5,)),
      ('2d probs, 1-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5,)),
      ('2d logits, 2-tuple shape', {'logits': [[0.0, 1.0, -0.5],
                                               [-0.1, 0.3, 0.0]]}, (5, 4)),
      ('2d probs, 2-tuple shape', {'probs': [[0.1, 0.4, 0.5],
                                             [0.5, 0.25, 0.25]]}, (5, 4)),
  )
  def test_sample_and_log_prob(self, distr_params, sample_shape):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_sample_and_log_prob(
        dist_args=(),
        dist_kwargs=distr_params,
        sample_shape=sample_shape,
        assertion_fn=self.assertion_fn(rtol=2e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('int32', jnp.int32),
      ('int64', jnp.int64),
      ('float32', jnp.float32),
      ('float64', jnp.float64))
  def test_sample_dtype(self, dtype):
    with jax.experimental.enable_x64(dtype.dtype.itemsize == 8):
      dist_params = {'logits': self.logits, 'dtype': dtype}
      dist = self.distrax_cls(**dist_params)
      samples = self.variant(dist.sample)(seed=self.key)
      self.assertEqual(samples.dtype, dist.dtype)
      chex.assert_type(samples, dtype)

  @chex.all_variants
  @parameterized.named_parameters(
      ('from probs', False),
      ('from logits', True))
  def test_sample_unique_values(self, from_logits):
    dist_params = {'logits': self.logits} if from_logits else {'probs': self.p}
    dist = self.distrax_cls(**dist_params)
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=100))
    samples = sample_fn(self.key)
    np.testing.assert_equal(np.unique(samples), np.arange(2))

  @chex.all_variants
  def test_sample_extreme_probs(self):
    dist_params = {'probs': np.asarray([1., 0., 0., 0.])}
    dist = self.distrax_cls(**dist_params)
    sample_fn = self.variant(
        lambda key: dist.sample(seed=key, sample_shape=100))
    samples = sample_fn(self.key)
    np.testing.assert_equal(np.unique(samples[..., 0]), 1)
    np.testing.assert_equal(np.unique(samples[..., 1:]), 0)

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob; 1d logits, 1 input',
       'log_prob',
       {'logits': [0.0, 0.5, -0.5]},
       [1, 0, 0]),
      ('log_prob; 1d logits, 2 inputs',
       'log_prob',
       {'logits': [0.0, 0.5, -0.5]},
       [[1, 0, 0], [0, 1, 0]]),
      ('log_prob; 2d logits, 2 inputs',
       'log_prob',
       {'logits': [[0.0, 0.5, -0.5], [-0.1, 0.1, 0.1]]},
       [[1, 0, 0], [0, 1, 0]]),
      ('log_prob; 2d logits, rank-3 inputs',
       'log_prob',
       {'logits': [[0.0, 0.5, -0.5], [-0.1, 0.1, 0.1]]},
       np.asarray([[1, 0, 0], [0, 1, 0]])[None, ...]),
      ('log_prob; 1d probs, 1 input',
       'log_prob',
       {'probs': [0.3, 0.2, 0.5]},
       [1, 0, 0]),
      ('log_prob; 1d probs, 2 inputs',
       'log_prob',
       {'probs': [0.3, 0.2, 0.5]},
       [[1, 0, 0], [0, 1, 0]]),
      ('log_prob; 2d probs, 2 inputs',
       'log_prob',
       {'probs': [[0.2, 0.4, 0.4], [0.1, 0.2, 0.7]]},
       [[1, 0, 0], [0, 1, 0]]),
      ('log_prob; 2d probs, rank-3 inputs',
       'log_prob',
       {'probs': [[0.2, 0.4, 0.4], [0.1, 0.2, 0.7]]},
       np.asarray([[1, 0, 0], [0, 1, 0]])[None, ...]),
      ('log_prob; unnormalized probs',
       'log_prob',
       {'probs': [0.1, 0.2, 0.3]},
       [[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
      ('prob; 1d logits, 1 input',
       'prob',
       {'logits': [0.0, 0.5, -0.5]},
       [1, 0, 0]),
      ('prob; 1d logits, 2 inputs',
       'prob',
       {'logits': [0.0, 0.5, -0.5]},
       [[1, 0, 0], [0, 1, 0]]),
      ('prob; 2d logits, 2 inputs',
       'prob',
       {'logits': [[0.0, 0.5, -0.5], [-0.1, 0.1, 0.1]]},
       [[1, 0, 0], [0, 1, 0]]),
      ('prob; 2d logits, rank-3 inputs',
       'prob',
       {'logits': [[0.0, 0.5, -0.5], [-0.1, 0.1, 0.1]]},
       np.asarray([[1, 0, 0], [0, 1, 0]])[None, ...]),
      ('prob; 1d probs, 1 input',
       'prob',
       {'probs': [0.3, 0.2, 0.5]},
       [1, 0, 0]),
      ('prob; 1d probs, 2 inputs',
       'prob',
       {'probs': [0.3, 0.2, 0.5]},
       [[1, 0, 0], [0, 1, 0]]),
      ('prob; 2d probs, 2 inputs',
       'prob',
       {'probs': [[0.2, 0.4, 0.4], [0.1, 0.2, 0.7]]},
       [[1, 0, 0], [0, 1, 0]]),
      ('prob; 2d probs, rank-3 inputs',
       'prob',
       {'probs': [[0.2, 0.4, 0.4], [0.1, 0.2, 0.7]]},
       np.asarray([[1, 0, 0], [0, 1, 0]])[None, ...]),
      ('prob; unnormalized probs',
       'prob',
       {'probs': [0.1, 0.2, 0.3]},
       [[0, 0, 1], [0, 1, 0], [1, 0, 0]]),
  )
  def test_pdf(self, function_string, distr_params, value):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    value = np.array(value)
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=distr_params,
        call_args=(value,),
        assertion_fn=self.assertion_fn(rtol=2e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('log_prob; extreme probs',
       'log_prob',
       {'probs': [0.0, 1.0, 0.0]},
       [[0, 1, 0], [1, 0, 0]],
       np.asarray([0., -np.inf])),
      ('prob; extreme probs',
       'prob',
       {'probs': [0.0, 1.0, 0.0]},
       [[0, 1, 0], [1, 0, 0]],
       np.asarray([1., 0.])),
  )
  def test_pdf_extreme_probs(self, function_string, distr_params,
                             value, expected):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    value = np.array(value)
    dist = self.distrax_cls(**distr_params)
    self.assertion_fn(rtol=2e-3)(
        self.variant(getattr(dist, function_string))(value), expected)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('entropy; from 2d logits',
       'entropy', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]}),
      ('entropy; from 2d probs',
       'entropy', {'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]]}),
      ('mode; from 2d logits',
       'mode', {'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]]}),
      ('mode; from 2d probs',
       'mode', {'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]]}),
  )
  def test_method(self, function_string, distr_params):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    super()._test_attribute(
        attribute_string=function_string,
        dist_kwargs=distr_params,
        call_args=(),
        assertion_fn=self.assertion_fn(rtol=2e-3))

  @chex.all_variants
  @parameterized.named_parameters(
      ('from 2d logits', {
          'logits': [[0.0, 0.5, -0.5], [-0.2, 0.3, 0.5]],
      }, [[0, 1, 0], [1, 0, 0]]),
      ('from 2d probs', {
          'probs': [[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]],
      }, [[0, 1, 0], [1, 0, 0]]),
  )
  def test_cdf(self, distr_params, values):
    distr_params = {k: jnp.asarray(v) for k, v in distr_params.items()}
    values = np.array(values)
    dist = self.distrax_cls(**distr_params)
    if 'probs' in distr_params:
      probs = distr_params['probs']
    else:
      probs = scipy.special.softmax(distr_params['logits'], axis=-1)
    expected = np.sum(np.cumsum(probs, axis=-1) * values, axis=-1)
    self.assertion_fn(rtol=2e-3)(self.variant(dist.cdf)(values), expected)

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_two_distributions(self, function_string, mode_string):
    super()._test_with_two_distributions(
        attribute_string=function_string,
        mode_string=mode_string,
        dist1_kwargs={'probs': np.array([[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]])},
        dist2_kwargs={'logits': np.array([0.0, 0.1, 0.1]),},
        assertion_fn=self.assertion_fn(rtol=2e-3))

  @chex.all_variants(with_pmap=False)
  @parameterized.named_parameters(
      ('kl distrax_to_distrax', 'kl_divergence', 'distrax_to_distrax'),
      ('kl distrax_to_tfp', 'kl_divergence', 'distrax_to_tfp'),
      ('kl tfp_to_distrax', 'kl_divergence', 'tfp_to_distrax'),
      ('cross-ent distrax_to_distrax', 'cross_entropy', 'distrax_to_distrax'),
      ('cross-ent distrax_to_tfp', 'cross_entropy', 'distrax_to_tfp'),
      ('cross-ent tfp_to_distrax', 'cross_entropy', 'tfp_to_distrax'))
  def test_with_categorical_and_one_hot_categorical(
      self, function_string, mode_string):
    dist1_params = {'probs': np.array([[0.1, 0.5, 0.4], [0.2, 0.4, 0.4]])}
    dist2_params = {'logits': np.array([0.0, 0.1, 0.1]),}

    dist1 = categorical.Categorical(**dist1_params)
    tfp_dist1 = tfd.Categorical(**dist1_params)
    dist2 = one_hot_categorical.OneHotCategorical(**dist2_params)
    tfp_dist2 = tfd.OneHotCategorical(**dist2_params)

    distrax_fn_1 = self.variant(getattr(dist1, function_string))
    distrax_fn_2 = self.variant(getattr(dist2, function_string))

    if mode_string == 'distrax_to_distrax':
      comp_dist1_dist2 = distrax_fn_1(dist2)
      comp_dist2_dist1 = distrax_fn_2(dist1)
    elif mode_string == 'distrax_to_tfp':
      comp_dist1_dist2 = distrax_fn_1(tfp_dist2)
      comp_dist2_dist1 = distrax_fn_2(tfp_dist1)
    elif mode_string == 'tfp_to_distrax':
      comp_dist1_dist2 = getattr(tfp_dist1, function_string)(dist2)
      comp_dist2_dist1 = getattr(tfp_dist2, function_string)(dist1)
    else:
      raise ValueError(f'Unsupported mode: {mode_string}')

    # The target values (obtained with TFP-only methods) are obtained with two
    # distributions of the same class (namely, Categorical) because TFP doesn't
    # register KLs of the form KL(Categorical || OneHotCategorical).
    tfp_dist2_aux = tfd.Categorical(**dist2_params)
    tfp_comp_dist1_dist2 = getattr(tfp_dist1, function_string)(tfp_dist2_aux)
    tfp_comp_dist2_dist1 = getattr(tfp_dist2_aux, function_string)(tfp_dist1)

    self.assertion_fn(rtol=2e-3)(comp_dist1_dist2, tfp_comp_dist1_dist2)
    self.assertion_fn(rtol=2e-3)(comp_dist2_dist1, tfp_comp_dist2_dist1)

  def test_jittable(self):
    super()._test_jittable((np.zeros((3,)),))

  @parameterized.named_parameters(
      ('single element', 2),
      ('range', slice(-1)),
      ('range_2', (slice(None), slice(-1))),
  )
  def test_slice(self, slice_):
    logits = jnp.array(np.random.randn(3, 4, 5))
    probs = jax.nn.softmax(jnp.array(np.random.randn(3, 4, 5)), axis=-1)
    dist1 = self.distrax_cls(logits=logits)
    dist2 = self.distrax_cls(probs=probs)
    dist1_sliced = dist1[slice_]
    dist2_sliced = dist2[slice_]
    self.assertion_fn(rtol=2e-3)(
        jax.nn.softmax(dist1_sliced.logits, axis=-1),
        jax.nn.softmax(logits[slice_], axis=-1))
    self.assertion_fn(rtol=2e-3)(dist2_sliced.probs, probs[slice_])
    self.assertIsInstance(dist1_sliced, one_hot_categorical.OneHotCategorical)
    self.assertIsInstance(dist2_sliced, one_hot_categorical.OneHotCategorical)

  def test_slice_ellipsis(self):
    logits = jnp.array(np.random.randn(4, 4, 5))
    probs = jax.nn.softmax(jnp.array(np.random.randn(4, 4, 5)), axis=-1)
    dist1 = self.distrax_cls(logits=logits)
    dist2 = self.distrax_cls(probs=probs)
    self.assertion_fn(rtol=2e-3)(
        jax.nn.softmax(dist1[..., -1].logits, axis=-1),
        jax.nn.softmax(logits[:, -1], axis=-1))
    self.assertion_fn(rtol=2e-3)(dist2[..., -1].probs, probs[:, -1])


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Beta distribution."""

from typing import Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
from distrax._src.utils import math
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Beta(distribution.Distribution):
  """Beta distribution with parameters `alpha` and `beta`.

  The PDF of a Beta distributed random variable `X` is defined on the interval
  `0 <= X <= 1` and has the form:
  ```
  p(x; alpha, beta) = x ** {alpha - 1} * (1 - x) ** (beta - 1) / B(alpha, beta)
  ```
  where `B(alpha, beta)` is the beta function, and the `alpha, beta > 0` are the
  shape parameters.

  Note that the support of the distribution does not include `x = 0` or `x = 1`
  if `alpha < 1` or `beta < 1`, respectively.
  """

  equiv_tfp_cls = tfd.Beta

  def __init__(self, alpha: Numeric, beta: Numeric):
    """Initializes a Beta distribution.

    Args:
      alpha: Shape parameter `alpha` of the distribution. Must be positive.
      beta: Shape parameter `beta` of the distribution. Must be positive.
    """
    super().__init__()
    self._alpha = conversion.as_float_array(alpha)
    self._beta = conversion.as_float_array(beta)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._alpha.shape, self._beta.shape)
    self._log_normalization_constant = math.log_beta(self._alpha, self._beta)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  @property
  def alpha(self) -> Array:
    """Shape parameter `alpha` of the distribution."""
    return jnp.broadcast_to(self._alpha, self.batch_shape)

  @property
  def beta(self) -> Array:
    """Shape parameter `beta` of the distribution."""
    return jnp.broadcast_to(self._beta, self.batch_shape)

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    out_shape = (n,) + self.batch_shape
    dtype = jnp.result_type(self._alpha, self._beta)
    rnd = jax.random.beta(
        key, a=self._alpha, b=self._beta, shape=out_shape, dtype=dtype)
    return rnd

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    result = ((self._alpha - 1.) * jnp.log(value)
              + (self._beta - 1.) * jnp.log(1. - value)
              - self._log_normalization_constant)
    return jnp.where(
        jnp.logical_or(jnp.logical_and(self._alpha == 1., value == 0.),
                       jnp.logical_and(self._beta == 1., value == 1.)),
        -self._log_normalization_constant,
        result
    )

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    return jax.scipy.special.betainc(self._alpha, self._beta, value)

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    return (
        self._log_normalization_constant
        - (self._alpha - 1.) * jax.lax.digamma(self._alpha)
        - (self._beta - 1.) * jax.lax.digamma(self._beta)
        + (self._alpha + self._beta - 2.) * jax.lax.digamma(
            self._alpha + self._beta)
    )

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._alpha / (self._alpha + self._beta)

  def variance(self) -> Array:
    """Calculates the variance."""
    sum_alpha_beta = self._alpha + self._beta
    return self._alpha * self._beta / (
        jnp.square(sum_alpha_beta) * (sum_alpha_beta + 1.))

  def mode(self) -> Array:
    """Calculates the mode.

    Returns:
      The mode, an array of shape `batch_shape`. The mode is not defined if
      `alpha = beta = 1`, or if `alpha < 1` and `beta < 1`. For these cases,
      the returned value is `jnp.nan`.
    """
    return jnp.where(
        jnp.logical_and(self._alpha > 1., self._beta > 1.),
        (self._alpha - 1.) / (self._alpha + self._beta - 2.),
        jnp.where(
            jnp.logical_and(self._alpha <= 1., self._beta > 1.),
            0.,
            jnp.where(
                jnp.logical_and(self._alpha > 1., self._beta <= 1.),
                1., jnp.nan)))

  def __getitem__(self, index) -> 'Beta':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Beta(alpha=self.alpha[index], beta=self.beta[index])


BetaLike = Union[Beta, tfd.Beta]


def _obtain_alpha_beta(dist: BetaLike) -> Tuple[Array, Array]:
  if isinstance(dist, Beta):
    alpha, beta = dist.alpha, dist.beta
  elif isinstance(dist, tfd.Beta):
    alpha, beta = dist.concentration1, dist.concentration0
  else:
    raise ValueError(f'Unsupported distribution: {type(dist)}.')
  return alpha, beta


def _kl_divergence_beta_beta(
    dist1: BetaLike,
    dist2: BetaLike,
    *unused_args,
    **unused_kwargs,
) -> Array:
  """Batched KL divergence KL(dist1 || dist2) between two Beta distributions.

  Args:
    dist1: A Beta distribution.
    dist2: A Beta distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  alpha1, beta1 = _obtain_alpha_beta(dist1)
  alpha2, beta2 = _obtain_alpha_beta(dist2)
  t1 = math.log_beta(alpha2, beta2) - math.log_beta(alpha1, beta1)
  t2 = (alpha1 - alpha2) * jax.lax.digamma(alpha1)
  t3 = (beta1 - beta2) * jax.lax.digamma(beta1)
  t4 = (alpha2 - alpha1 + beta2 - beta1) * jax.lax.digamma(alpha1 + beta1)
  return t1 + t2 + t3 + t4


# Register the KL functions with TFP.
tfd.RegisterKL(Beta, Beta)(_kl_divergence_beta_beta)
tfd.RegisterKL(Beta, Beta.equiv_tfp_cls)(_kl_divergence_beta_beta)
tfd.RegisterKL(Beta.equiv_tfp_cls, Beta)(_kl_divergence_beta_beta)



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tests for `transformations.py`."""

from absl.testing import absltest
from absl.testing import parameterized

import chex
from distrax._src.utils import transformations
import jax
import jax.numpy as jnp
import numpy as np
from tensorflow_probability.substrates import jax as tfp


tfd = tfp.distributions
tfb = tfp.bijectors


RTOL = 1e-2


class TransformationsTest(parameterized.TestCase):

  def setUp(self):
    super().setUp()
    self.seed = jax.random.PRNGKey(1234)

  @chex.all_variants
  @parameterized.named_parameters(
      ('tanh', jnp.tanh, jnp.arctanh, 0.5),
      ('arctanh', jnp.arctanh, jnp.tanh, 0.5),
      ('sinh', jnp.sinh, jnp.arcsinh, 0.5),
      ('arcsinh', jnp.arcsinh, jnp.sinh, 0.5),
      ('cosh', jnp.cosh, jnp.arccosh, 0.5),
      ('arccosh', jnp.arccosh, jnp.cosh, 2.0),
      ('exp', jnp.exp, jnp.log, 0.5),
      ('log', jnp.log, jnp.exp, 0.5),
      ('pow', lambda x: jnp.power(x, 3.5), lambda y: jnp.power(y, 1/3.5), 0.5),
      ('add', lambda x: x + 3, lambda y: y - 3, 0.5),
      ('sqrt', jnp.sqrt, jnp.square, 0.5),
      ('square', jnp.square, jnp.sqrt, 0.5),
      ('reciprocal', jnp.reciprocal, jnp.reciprocal, 0.5),
      ('negate', lambda x: -x, lambda y: -y, 0.5),
      ('log1p', jnp.log1p, jnp.expm1, 0.5),
      ('expm1', jnp.expm1, jnp.log1p, 0.5),
      ('erf', jax.lax.erf, jax.lax.erf_inv, 0.5),
      ('erf_inv', jax.lax.erf_inv, jax.lax.erf, 0.5),
      ('2_mul_x', lambda x: 2 * x, lambda y: y * 0.5, 0.5),
      ('x_mul_2', lambda x: x * 2, lambda y: y * 0.5, 0.5),
      ('2_div_x', lambda x: 2 / x, lambda y: 2 / y, 0.5),
      ('x_div_2', lambda x: x / 2, lambda y: y / 0.5, 0.5),
      ('x_sub_3', lambda x: x - 3, lambda y: y + 3, 0.5),
      ('3_sub_x', lambda x: 3 - x, lambda y: 3 - y, 0.5),
      ('x**3.5', lambda x: x**3.5, lambda y: y**(1/3.5), 0.5),
      ('x**(1/3.5)', lambda x: x**(1/3.5), lambda y: y**3.5, 0.5),
  )
  def test_inversion(self, forward, inverse, x):
    x = jnp.array([x], dtype=jnp.float32)
    y = forward(x)

    inverse_ = self.variant(transformations.inv(forward))
    x_ = inverse_(y)

    np.testing.assert_allclose(x_, x, rtol=RTOL)
    np.testing.assert_allclose(x_, inverse(y), rtol=RTOL)

  @chex.all_variants
  @parameterized.named_parameters(
      ('tanh', jnp.tanh, jnp.arctanh, 0.5),
      ('arctanh', jnp.arctanh, jnp.tanh, 0.5),
      ('sinh', jnp.sinh, jnp.arcsinh, 0.5),
      ('arcsinh', jnp.arcsinh, jnp.sinh, 0.5),
      ('cosh', jnp.cosh, jnp.arccosh, 0.5),
      ('arccosh', jnp.arccosh, jnp.cosh, 2.0),
      ('exp', jnp.exp, jnp.log, 0.5),
      ('log', jnp.log, jnp.exp, 0.5),
      ('pow', lambda x: jnp.power(x, 3.5), lambda y: jnp.power(y, 1/3.5), 0.5),
      ('add', lambda x: x + 3, lambda y: y - 3, 0.5),
      ('sqrt', jnp.sqrt, jnp.square, 0.5),
      ('square', jnp.square, jnp.sqrt, 0.5),
      ('reciprocal', jnp.reciprocal, jnp.reciprocal, 0.5),
      ('negate', lambda x: -x, lambda y: -y, 0.5),
      ('log1p', jnp.log1p, jnp.expm1, 0.5),
      ('expm1', jnp.expm1, jnp.log1p, 0.5),
      ('erf', jax.lax.erf, jax.lax.erf_inv, 0.5),
      ('erf_inv', jax.lax.erf_inv, jax.lax.erf, 0.5),
      ('2_mul_x', lambda x: 2 * x, lambda y: y * 0.5, 0.5),
      ('x_mul_2', lambda x: x * 2, lambda y: y * 0.5, 0.5),
      ('2_div_x', lambda x: 2 / x, lambda y: 2 / y, 0.5),
      ('x_div_2', lambda x: x / 2, lambda y: y / 0.5, 0.5),
      ('x_sub_3', lambda x: x - 3, lambda y: y + 3, 0.5),
      ('3_sub_x', lambda x: 3 - x, lambda y: 3 - y, 0.5),
      ('x**3.5', lambda x: x**3.5, lambda y: y**(1/3.5), 0.5),
      ('x**(1/3.5)', lambda x: x**(1/3.5), lambda y: y**3.5, 0.5),
  )
  def test_inverting_jitted_function(self, forward, inverse, x):
    x = jnp.array([x], dtype=jnp.float32)
    y = forward(x)

    jitted_forward = jax.jit(forward)
    inverse_ = self.variant(transformations.inv(jitted_forward))
    x_ = inverse_(y)

    np.testing.assert_allclose(x_, x, rtol=RTOL)
    np.testing.assert_allclose(x_, inverse(y), rtol=RTOL)

  @chex.all_variants
  @parameterized.named_parameters(
      ('identity, 0d', lambda x: x, tfb.Identity, 0.5),
      ('identity, 1d', lambda x: x, tfb.Identity, [0.9]),
      ('identity, 2d', lambda x: x, tfb.Identity, [0.25, 0.75]),
      ('identity, 2x2d', lambda x: x, tfb.Identity, [[0.25, 0.75],
                                                     [0.1, 0.9]]),
      ('scale, 0d', lambda x: 3.0 * x, lambda: tfb.Scale(3.0), 0.5),
      ('scale, 1d', lambda x: 3.0 * x, lambda: tfb.Scale(3.0), [0.9]),
      ('scale, 2d', lambda x: 3.0 * x, lambda: tfb.Scale(3.0), [0.25, 0.75]),
      ('scale, 2x2d', lambda x: 3.0 * x, lambda: tfb.Scale(3.0), [[0.25, 0.75],
                                                                  [0.1, 0.9]]),
      ('tanh, 0d', jnp.tanh, tfb.Tanh, 0.5),
      ('tanh, 1d', jnp.tanh, tfb.Tanh, [0.9]),
      ('tanh, 2d', jnp.tanh, tfb.Tanh, [0.25, 0.75]),
      ('tanh, 2x2d', jnp.tanh, tfb.Tanh, [[0.25, 0.75],
                                          [0.1, 0.9]]),
      ('softplus, 0d', jax.nn.softplus, tfb.Softplus, 0.5),
      ('softplus, 1d', jax.nn.softplus, tfb.Softplus, [0.9]),
      ('softplus, 2d', jax.nn.softplus, tfb.Softplus, [0.25, 0.75]),
      ('softplus, 2x2d', jax.nn.softplus, tfb.Softplus, [[0.25, 0.75],
                                                         [0.1, 0.9]]),
      ('sigmoid, 0d', jax.nn.sigmoid, tfb.Sigmoid, 0.5),
      ('sigmoid, 1d', jax.nn.sigmoid, tfb.Sigmoid, [0.9]),
      ('sigmoid, 2d', jax.nn.sigmoid, tfb.Sigmoid, [0.25, 0.75]),
      ('sigmoid, 2x2d', jax.nn.sigmoid, tfb.Sigmoid, [[0.25, 0.75],
                                                      [0.1, 0.9]]),
  )
  def test_log_det_scalar(self, forward, tfb_bijector, x):
    x = np.array(x, dtype=np.float32)

    log_det_fn = self.variant(transformations.log_det_scalar(forward))

    actual = log_det_fn(x)
    expected = tfb_bijector().forward_log_det_jacobian(x, event_ndims=0)

    np.testing.assert_allclose(actual, expected, rtol=RTOL)

  @parameterized.named_parameters(
      ('tanh', jnp.tanh, False),
      ('sigmoid', jax.nn.sigmoid, False),
      ('identity', lambda x: x, True),
      ('square', lambda x: x**2, False),
      ('softplus', jax.nn.softplus, False),
      ('exp', jnp.exp, False),
      ('log', jnp.log, False, 1.0),
      ('shift', lambda x: x + 3.0, True),
      ('scale', lambda x: 2.0 * x, True),
      ('shift and scale', lambda x: 2.0 * x + 3.0, True),
  )
  def test_is_constant_jacobian(self, fn, is_constant, x=0.0):
    is_constant_ = transformations.is_constant_jacobian(fn, x)
    np.testing.assert_array_equal(is_constant, is_constant_)


if __name__ == '__main__':
  absltest.main()



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Wrapper to turn independent Bijectors into block Bijectors."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
from distrax._src.utils import conversion
from distrax._src.utils import math

Array = base.Array
BijectorLike = base.BijectorLike
BijectorT = base.BijectorT


class Block(base.Bijector):
  """A wrapper that promotes a bijector to a block bijector.

  A block bijector applies a bijector to a k-dimensional array of events, but
  considers that array of events to be a single event. In practical terms, this
  means that the log det Jacobian will be summed over its last k dimensions.

  For example, consider a scalar bijector (such as `Tanh`) that operates on
  scalar events. We may want to apply this bijector identically to a 4D array of
  shape [N, H, W, C] representing a sequence of N images. Doing so naively will
  produce a log det Jacobian of shape [N, H, W, C], because the scalar bijector
  will assume scalar events and so all 4 dimensions will be considered as batch.
  To promote the scalar bijector to a "block scalar" that operates on the 3D
  arrays can be done by `Block(bijector, ndims=3)`. Then, applying the block
  bijector will produce a log det Jacobian of shape [N] as desired.

  In general, suppose `bijector` operates on n-dimensional events. Then,
  `Block(bijector, k)` will promote `bijector` to a block bijector that
  operates on (k + n)-dimensional events, summing the log det Jacobian over its
  last k dimensions. In practice, this means that the last k batch dimensions
  will be turned into event dimensions.
  """

  def __init__(self, bijector: BijectorLike, ndims: int):
    """Initializes a Block.

    Args:
      bijector: the bijector to be promoted to a block bijector. It can be a
        distrax bijector, a TFP bijector, or a callable to be wrapped by
        `Lambda`.
      ndims: number of batch dimensions to promote to event dimensions.
    """
    if ndims < 0:
      raise ValueError(f"`ndims` must be non-negative; got {ndims}.")
    self._bijector = conversion.as_bijector(bijector)
    self._ndims = ndims
    super().__init__(
        event_ndims_in=ndims + self._bijector.event_ndims_in,
        event_ndims_out=ndims + self._bijector.event_ndims_out,
        is_constant_jacobian=self._bijector.is_constant_jacobian,
        is_constant_log_det=self._bijector.is_constant_log_det)

  @property
  def bijector(self) -> BijectorT:
    """The base bijector, without promoting to a block bijector."""
    return self._bijector

  @property
  def ndims(self) -> int:
    """The number of batch dimensions promoted to event dimensions."""
    return self._ndims

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    self._check_forward_input_shape(x)
    return self._bijector.forward(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    self._check_inverse_input_shape(y)
    return self._bijector.inverse(y)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    log_det = self._bijector.forward_log_det_jacobian(x)
    return math.sum_last(log_det, self._ndims)

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    self._check_inverse_input_shape(y)
    log_det = self._bijector.inverse_log_det_jacobian(y)
    return math.sum_last(log_det, self._ndims)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    self._check_forward_input_shape(x)
    y, log_det = self._bijector.forward_and_log_det(x)
    return y, math.sum_last(log_det, self._ndims)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    self._check_inverse_input_shape(y)
    x, log_det = self._bijector.inverse_and_log_det(y)
    return x, math.sum_last(log_det, self._ndims)

  @property
  def name(self) -> str:
    """Name of the bijector."""
    return self.__class__.__name__ + self._bijector.name

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is Block:  # pylint: disable=unidiomatic-typecheck
      return self.bijector.same_as(other.bijector) and self.ndims == other.ndims

    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Tanh bijector."""

from typing import Tuple

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp


Array = base.Array


class Tanh(base.Bijector):
  """A bijector that computes the hyperbolic tangent.

  The log-determinant implementation in this bijector is more numerically stable
  than relying on the automatic differentiation approach used by Lambda, so this
  bijector should be preferred over Lambda(jnp.tanh) where possible. See
  `tfp.bijectors.Tanh` for details.

  When the absolute value of the input is large, `Tanh` becomes close to a
  constant, so that it is not possible to recover the input `x` from the output
  `y` within machine precision. In cases where it is needed to compute both the
  forward mapping and the backward mapping one after the other to recover the
  original input `x`, it is the user's responsibility to simplify the operation
  to avoid numerical issues; this is unlike the `tfp.bijectors.Tanh`. One
  example of such case is to use the bijector within a `Transformed`
  distribution and to obtain the log-probability of samples obtained from the
  distribution's `sample` method. For values of the samples for which it is not
  possible to apply the inverse bijector accurately, `log_prob` returns NaN.
  This can be avoided by using `sample_and_log_prob` instead of `sample`
  followed by `log_prob`.
  """

  def __init__(self):
    """Initializes a Tanh bijector."""
    super().__init__(event_ndims_in=0)

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    return 2 * (jnp.log(2) - x - jax.nn.softplus(-2 * x))

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return jnp.tanh(x), self.forward_log_det_jacobian(x)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    x = jnp.arctanh(y)
    return x, -self.forward_log_det_jacobian(x)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    return type(other) is Tanh  # pylint: disable=unidiomatic-typecheck



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Shift bijector."""

from typing import Tuple, Union

from distrax._src.bijectors import bijector as base
import jax
import jax.numpy as jnp

Array = base.Array
Numeric = Union[Array, float]


class Shift(base.Bijector):
  """Bijector that translates its input elementwise.

  The bijector is defined as follows:

  - Forward: `y = x + shift`
  - Forward Jacobian determinant: `log|det J(x)| = 0`
  - Inverse: `x = y - shift`
  - Inverse Jacobian determinant: `log|det J(y)| = 0`

  where `shift` parameterizes the bijector.
  """

  def __init__(self, shift: Numeric):
    """Initializes a `Shift` bijector.

    Args:
      shift: the bijector's shift parameter. Can also be batched.
    """
    super().__init__(event_ndims_in=0, is_constant_jacobian=True)
    self._shift = shift
    self._batch_shape = jnp.shape(self._shift)

  @property
  def shift(self) -> Numeric:
    """The bijector's shift."""
    return self._shift

  def forward(self, x: Array) -> Array:
    """Computes y = f(x)."""
    return x + self._shift

  def forward_log_det_jacobian(self, x: Array) -> Array:
    """Computes log|det J(f)(x)|."""
    batch_shape = jax.lax.broadcast_shapes(self._batch_shape, x.shape)
    return jnp.zeros(batch_shape, dtype=x.dtype)

  def forward_and_log_det(self, x: Array) -> Tuple[Array, Array]:
    """Computes y = f(x) and log|det J(f)(x)|."""
    return self.forward(x), self.forward_log_det_jacobian(x)

  def inverse(self, y: Array) -> Array:
    """Computes x = f^{-1}(y)."""
    return y - self._shift

  def inverse_log_det_jacobian(self, y: Array) -> Array:
    """Computes log|det J(f^{-1})(y)|."""
    return self.forward_log_det_jacobian(y)

  def inverse_and_log_det(self, y: Array) -> Tuple[Array, Array]:
    """Computes x = f^{-1}(y) and log|det J(f^{-1})(y)|."""
    return self.inverse(y), self.inverse_log_det_jacobian(y)

  def same_as(self, other: base.Bijector) -> bool:
    """Returns True if this bijector is guaranteed to be the same as `other`."""
    if type(other) is Shift:  # pylint: disable=unidiomatic-typecheck
      return self.shift is other.shift
    return False



# Copyright 2021 DeepMind Technologies Limited. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Gamma distribution."""

from typing import Tuple, Union

import chex
from distrax._src.distributions import distribution
from distrax._src.utils import conversion
import jax
import jax.numpy as jnp
from tensorflow_probability.substrates import jax as tfp

tfd = tfp.distributions

Array = chex.Array
Numeric = chex.Numeric
PRNGKey = chex.PRNGKey
EventT = distribution.EventT


class Gamma(distribution.Distribution):
  """Gamma distribution with parameters `concentration` and `rate`."""

  equiv_tfp_cls = tfd.Gamma

  def __init__(self, concentration: Numeric, rate: Numeric):
    """Initializes a Gamma distribution.

    Args:
      concentration: Concentration parameter of the distribution.
      rate: Inverse scale params of the distribution.
    """
    super().__init__()
    self._concentration = conversion.as_float_array(concentration)
    self._rate = conversion.as_float_array(rate)
    self._batch_shape = jax.lax.broadcast_shapes(
        self._concentration.shape, self._rate.shape)

  @property
  def event_shape(self) -> Tuple[int, ...]:
    """Shape of event of distribution samples."""
    return ()

  @property
  def batch_shape(self) -> Tuple[int, ...]:
    """Shape of batch of distribution samples."""
    return self._batch_shape

  @property
  def concentration(self) -> Array:
    """Concentration of the distribution."""
    return jnp.broadcast_to(self._concentration, self.batch_shape)

  @property
  def rate(self) -> Array:
    """Inverse scale of the distribution."""
    return jnp.broadcast_to(self._rate, self.batch_shape)

  def _sample_from_std_gamma(self, key: PRNGKey, n: int) -> Array:
    out_shape = (n,) + self.batch_shape
    dtype = jnp.result_type(self._concentration, self._rate)
    return jax.random.gamma(
        key, a=self._concentration, shape=out_shape, dtype=dtype
    )

  def _sample_n(self, key: PRNGKey, n: int) -> Array:
    """See `Distribution._sample_n`."""
    rnd = self._sample_from_std_gamma(key, n)
    return rnd / self._rate

  def log_prob(self, value: EventT) -> Array:
    """See `Distribution.log_prob`."""
    return (
        self._concentration * jnp.log(self._rate)
        + (self._concentration - 1) * jnp.log(value)
        - self._rate * value
        - jax.lax.lgamma(self._concentration)
    )

  def entropy(self) -> Array:
    """Calculates the Shannon entropy (in nats)."""
    log_rate = jnp.log(self._rate)
    return (
        self._concentration
        - log_rate
        + jax.lax.lgamma(self._concentration)
        + (1.0 - self._concentration) * jax.lax.digamma(self._concentration)
    )

  def cdf(self, value: EventT) -> Array:
    """See `Distribution.cdf`."""
    return jax.lax.igamma(self._concentration, self._rate * value)

  def log_cdf(self, value: EventT) -> Array:
    """See `Distribution.log_cdf`."""
    return jnp.log(self.cdf(value))

  def mean(self) -> Array:
    """Calculates the mean."""
    return self._concentration / self._rate

  def stddev(self) -> Array:
    """Calculates the standard deviation."""
    return jnp.sqrt(self._concentration) / self._rate

  def variance(self) -> Array:
    """Calculates the variance."""
    return self._concentration / jnp.square(self._rate)

  def mode(self) -> Array:
    """Calculates the mode."""
    mode = (self._concentration - 1.0) / self._rate
    return jnp.where(self._concentration >= 1.0, mode, jnp.nan)

  def __getitem__(self, index) -> 'Gamma':
    """See `Distribution.__getitem__`."""
    index = distribution.to_batch_shape_index(self.batch_shape, index)
    return Gamma(
        concentration=self.concentration[index], rate=self.rate[index])


def _kl_divergence_gamma_gamma(
    dist1: Union[Gamma, tfd.Gamma],
    dist2: Union[Gamma, tfd.Gamma],
    *unused_args,
    **unused_kwargs,
) -> Array:
  """Batched KL divergence KL(dist1 || dist2) between two Gamma distributions.

  Args:
    dist1: A Gamma distribution.
    dist2: A Gamma distribution.

  Returns:
    Batchwise `KL(dist1 || dist2)`.
  """
  t1 = dist2.concentration * (jnp.log(dist1.rate) - jnp.log(dist2.rate))
  t2 = jax.lax.lgamma(dist2.concentration) - jax.lax.lgamma(dist1.concentration)
  t3 = (dist1.concentration - dist2.concentration) * jax.lax.digamma(
      dist1.concentration)
  t4 = (dist2.rate - dist1.rate) * (dist1.concentration / dist1.rate)
  return t1 + t2 + t3 + t4


# Register the KL functions with TFP.
tfd.RegisterKL(Gamma, Gamma)(_kl_divergence_gamma_gamma)
tfd.RegisterKL(Gamma, Gamma.equiv_tfp_cls)(_kl_divergence_gamma_gamma)
tfd.RegisterKL(Gamma.equiv_tfp_cls, Gamma)(_kl_divergence_gamma_gamma)
